import datetime
import logging
import pathlib
import sqlite3
import sys
from functools import lru_cache
from itertools import tee
from operator import itemgetter
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import OrderedDict
from typing import Set
from typing import Tuple
from typing import Union
from uuid import UUID
from uuid import uuid4

import click
import requests
from fastapi.encoders import jsonable_encoder
from more_itertools import last
from more_itertools import one
from more_itertools import partition
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_settings
from ramodels.mo import Employee
from ramodels.mo._shared import OrganisationRef
from tqdm import tqdm

from integrations import cpr_mapper
from integrations.ad_integration import ad_reader
from integrations.calculate_primary.common import LOGGER_NAME
from integrations.calculate_primary.common import NoPrimaryFound
from integrations.calculate_primary.sd import SDPrimaryEngagementUpdater
from integrations.rundb.db_overview import DBOverview
from integrations.SD_Lon import sd_payloads
from integrations.SD_Lon.config import ChangedAtSettings
from integrations.SD_Lon.config import get_changed_at_settings
from integrations.SD_Lon.date_utils import date_to_datetime
from integrations.SD_Lon.date_utils import gen_date_intervals
from integrations.SD_Lon.date_utils import sd_to_mo_termination_date
from integrations.SD_Lon.engagement import create_engagement
from integrations.SD_Lon.engagement import engagement_components
from integrations.SD_Lon.engagement import (
    is_employment_id_and_no_salary_minimum_consistent,
)
from integrations.SD_Lon.engagement import update_existing_engagement
from integrations.SD_Lon.fix_departments import FixDepartments
from integrations.SD_Lon.models import JobFunction
from integrations.SD_Lon.models import SDBasePerson
from integrations.SD_Lon.sd_common import calc_employment_id
from integrations.SD_Lon.sd_common import EmploymentStatus
from integrations.SD_Lon.sd_common import ensure_list
from integrations.SD_Lon.sd_common import mora_assert
from integrations.SD_Lon.sd_common import primary_types
from integrations.SD_Lon.sd_common import read_employment_at
from integrations.SD_Lon.sd_common import sd_lookup
from integrations.SD_Lon.sd_common import skip_fictional_users
from integrations.SD_Lon.sync_job_id import JobIdSync


LOG_LEVEL = logging.DEBUG
LOG_FILE = "mo_integrations.log"

logger = logging.getLogger("sdChangedAt")


def setup_logging():
    detail_logging = (
        "sdCommon",
        "sdChangedAt",
        LOGGER_NAME,
        "fixDepartments",
        "sdSyncJobId",
    )
    for name in logging.root.manager.loggerDict:
        if name in detail_logging:
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        filename=LOG_FILE,
    )


# TODO: SHOULD WE IMPLEMENT PREDICTABLE ENGAGEMENT UUIDS ALSO IN THIS CODE?!?


class ChangeAtSD:
    def __init__(
        self,
        settings: ChangedAtSettings,
        from_date: datetime.datetime,
        to_date: Optional[datetime.datetime] = None,
    ):
        self.settings = settings

        job_function_type = self.settings.sd_job_function
        if job_function_type == JobFunction.job_position_identifier:
            logger.info("Read settings. JobPositionIdentifier for job_functions")
            self.use_jpi = True
        elif job_function_type == JobFunction.employment_name:
            logger.info("Read settings. Do not update job_functions")
            self.use_jpi = False

        self.employee_forced_uuids = self._read_forced_uuids()
        self.department_fixer = self._get_fix_departments()
        self.helper = self._get_mora_helper(self.settings.mora_base)
        self.job_sync = self._get_job_sync(self.settings)

        # List of job_functions that should be ignored.
        self.skip_job_functions = self.settings.sd_skip_job_functions
        self.use_ad = self.settings.sd_use_ad_integration

        # TODO: remove when MO is updated to a version greater than 2.7.0
        self.terminate_engagement_with_to_only = (
            self.settings.sd_terminate_engagement_with_to_only
        )

        # See https://os2web.atlassian.net/browse/MO-245 for more details
        # about no_salary_minimum
        self.no_salary_minimum = self.settings.sd_no_salary_minimum_id

        try:
            self.org_uuid = self.helper.read_organisation()
        except requests.exceptions.RequestException as e:
            logger.error(e)
            print(e)
            exit()

        self.updater = self._get_primary_engagement_updater()
        self.from_date = from_date
        self.to_date = to_date

        # Cache of mo engagements
        self.mo_engagements_cache: Dict[str, list] = {}

        self.primary_types = self._get_primary_types(self.helper)

        logger.info("Read job_functions")
        facet_info = self.helper.read_classes_in_facet("engagement_job_function")
        job_functions = facet_info[0]
        self.job_function_facet = facet_info[1]
        # Map from user-key to uuid if jpi, name to uuid otherwise
        job_function_mapper = cast(
            Callable[[Any], Tuple[str, str]], itemgetter("name", "uuid")
        )
        if self.use_jpi:
            job_function_mapper = cast(
                Callable[[Any], Tuple[str, str]], itemgetter("user_key", "uuid")
            )
        self.job_functions: Dict[str, str] = dict(
            map(job_function_mapper, job_functions)
        )

        logger.info("Read engagement types")
        # The Opus diff-import contains a slightly more abstrac def to do this
        engagement_types = self.helper.read_classes_in_facet("engagement_type")
        self.engagement_type_facet = engagement_types[1]
        engagement_type_mapper = cast(
            Callable[[Any], Tuple[str, str]], itemgetter("user_key", "uuid")
        )
        self.engagement_types: Dict[str, str] = dict(
            map(engagement_type_mapper, engagement_types[0])
        )

        # SD supports only one type of leave
        self.leave_uuid = self.helper.ensure_class_in_facet("leave_type", "Orlov")

        facet_info = self.helper.read_classes_in_facet("association_type")
        self.association_uuid = facet_info[0][0]["uuid"]

    def _get_primary_types(self, mora_helper: MoraHelper):
        return primary_types(mora_helper)

    def _get_primary_engagement_updater(self) -> SDPrimaryEngagementUpdater:
        return SDPrimaryEngagementUpdater()

    def _get_fix_departments(self) -> FixDepartments:
        return FixDepartments(self.settings)

    def _get_mora_helper(self, mora_base) -> MoraHelper:
        return MoraHelper(hostname=mora_base, use_cache=False)

    def _get_job_sync(self, settings: ChangedAtSettings) -> JobIdSync:
        return JobIdSync(settings)

    def _read_forced_uuids(self):
        cpr_map = pathlib.Path.cwd() / "settings" / "cpr_uuid_map.csv"
        if not cpr_map.is_file():
            logger.error("Did not find cpr mapping")
            raise Exception("Did not find cpr mapping")

        logger.info("Found cpr mapping")
        employee_forced_uuids = cpr_mapper.employee_mapper(str(cpr_map))
        return employee_forced_uuids

    @lru_cache(maxsize=None)
    def _get_ad_reader(self):
        if self.use_ad:
            logger.info("AD integration in use")
            return ad_reader.ADParameterReader()
        logger.info("AD integration not in use")
        return None

    def _fetch_ad_information(self, cpr) -> Union[Tuple[None, None], Tuple[str, str]]:
        ad_reader = self._get_ad_reader()
        if ad_reader is None:
            return None, None

        ad_info = ad_reader.read_user(cpr=cpr)
        object_guid = ad_info.get("ObjectGuid", None)
        sam_account_name = ad_info.get("SamAccountName", None)
        return sam_account_name, object_guid

    @lru_cache(maxsize=None)
    def _fetch_ad_it_system_uuid(self):
        if not self.use_ad:
            raise ValueError("_fetch_ad_it_system_uuid called without AD enabled")
        it_systems = self.helper.read_it_systems()
        return one(
            map(
                itemgetter("uuid"),
                filter(lambda system: system["name"] == "Active Directory", it_systems),
            )
        )

    @lru_cache(maxsize=None)
    def read_employment_changed(
        self,
        from_date: Optional[datetime.datetime] = None,
        to_date: Optional[datetime.datetime] = None,
        employment_identifier: Optional[str] = None,
        in_cpr: Optional[str] = None,
    ):
        from_date = from_date or self.from_date
        to_date = to_date or self.to_date

        params = {
            "ActivationDate": from_date.strftime("%d.%m.%Y"),
            "ActivationTime": from_date.strftime("%H:%M"),
            "DepartmentIndicator": "true",
            "EmploymentStatusIndicator": "true",
            "ProfessionIndicator": "true",
            "WorkingTimeIndicator": "true",
            "UUIDIndicator": "true",
            "StatusPassiveIndicator": "true",
            "SalaryAgreementIndicator": "false",
            "SalaryCodeGroupIndicator": "false",
        }
        if employment_identifier:
            params.update(
                {
                    "EmploymentIdentifier": employment_identifier,
                }
            )
        if in_cpr:
            params.update(
                {
                    "PersonCivilRegistrationIdentifier": in_cpr,
                }
            )

        if to_date is not None:
            url = "GetEmploymentChangedAtDate20111201"
            params.update(
                {
                    "DeactivationDate": to_date.strftime("%d.%m.%Y"),
                    "DeactivationTime": to_date.strftime("%H:%M"),
                    "StatusActiveIndicator": "true",
                    "StatusPassiveIndicator": "true",
                    "FutureInformationIndicator": "true",
                }
            )
        else:
            url = "GetEmploymentChanged20111201"
            params.update(
                {
                    "DeactivationDate": "31.12.9999",
                }
            )
        response = sd_lookup(url, settings=self.settings, params=params)

        employment_response = ensure_list(response.get("Person", []))

        return employment_response

    def get_sd_persons_changed(
        self, from_date: datetime.datetime, to_date: Optional[datetime.datetime] = None
    ) -> List[OrderedDict[str, Any]]:
        """
        Get list of SD Løn persons that have changed between `from_date`
        and `to_date`

        Returns:
            List of SD Løn persons changed between the two dates
        """

        params = {
            "ActivationDate": from_date.strftime("%d.%m.%Y"),
            "ActivationTime": from_date.strftime("%H:%M"),
            "DeactivationDate": "31.12.9999",
            "StatusActiveIndicator": "true",
            "StatusPassiveIndicator": "true",
            "ContactInformationIndicator": "false",
            "PostalAddressIndicator": "false"
            # TODO: Er der kunder, som vil udlæse adresse-information?
        }
        if to_date:
            params["DeactivationDate"] = to_date.strftime("%d.%m.%Y")
            params["DeactivationTime"] = to_date.strftime("%H:%M")

        url = "GetPersonChangedAtDate20111201"
        response = sd_lookup(url, settings=self.settings, params=params)
        persons_changed = ensure_list(response.get("Person", []))
        return persons_changed

    def get_sd_person(self, cpr: str) -> List[OrderedDict[str, Any]]:
        """
        Get a single person from SD Løn at `self.from_date`

        Args:
            cpr: the cpr number of the person

        Returns:
            A list containing the person (or an empty list if no person
            is found)
        """

        params = {
            "EffectiveDate": self.from_date.strftime("%d.%m.%Y"),
            "PersonCivilRegistrationIdentifier": cpr,
            "StatusActiveIndicator": "True",
            "StatusPassiveIndicator": "false",
            "ContactInformationIndicator": "false",
            "PostalAddressIndicator": "false",
        }
        url = "GetPerson20111201"
        response = sd_lookup(url, settings=self.settings, params=params)
        person = ensure_list(response.get("Person", []))
        return person

    def update_changed_persons(
        self, in_cpr: Optional[str] = None, dry_run: bool = False
    ) -> None:
        """Update and insert (upsert) changed persons.

        Args:
            in_cpr: Optional CPR number of a specific person to upsert instead of
                    using SDs GetPersonChangedAtDate endpoint.

        Note:
            This method does not create employments at all, as this responsibility is
            handled by the update_employment method instead.
        """

        def extract_cpr_and_name(person: Dict[str, Any]) -> SDBasePerson:
            """
            Get CPR, given name (firstname) and surname (lastname) of
            SD Løn person.

            Args:
                person: the person from SD Løn

            Returns:
                Basic SD person with cpr and names
            """

            return SDBasePerson(
                cpr=person["PersonCivilRegistrationIdentifier"],
                given_name=person.get("PersonGivenName"),
                surname=person.get("PersonSurnameName"),
            )

        def fetch_mo_person(person: SDBasePerson) -> Dict[str, Any]:
            mo_person = self.helper.read_user(
                user_cpr=person.cpr, org_uuid=self.org_uuid
            )
            return mo_person

        def upsert_employee(
            uuid: str, given_name: Optional[str], sur_name: Optional[str], cpr: str
        ) -> str:
            model = Employee(
                uuid=uuid,
                user_key=uuid,
                givenname=given_name,
                surname=sur_name,
                cpr_no=cpr,
                org=OrganisationRef(uuid=self.org_uuid),
            )
            payload = jsonable_encoder(model.dict(by_alias=True, exclude_none=True))
            if dry_run:
                print("Dry-run: upsert_employee", payload)
                return "invalid-uuid"
            response = self.helper._mo_post("e/create", payload)
            assert response.status_code == 201
            return_uuid = response.json()
            logger.info(
                "Created or updated employee {} {} with uuid {}".format(
                    given_name, sur_name, return_uuid
                )
            )
            return return_uuid

        def create_itsystem_connection(sam_account_name: str, user_uuid: str):
            payload = sd_payloads.connect_it_system_to_user(
                sam_account_name, self._fetch_ad_it_system_uuid(), user_uuid
            )
            if dry_run:
                print("Dry-run: create_itsystem_connection", payload)
                return
            logger.debug("Connect it-system: {}".format(payload))
            response = self.helper._mo_post("details/create", payload)
            assert response.status_code == 201
            logger.info("Added AD account info to {}".format(user_uuid))

        # Fetch a list of persons to update
        if in_cpr is not None:
            all_sd_persons_changed = self.get_sd_person(in_cpr)
        else:
            logger.info("Update all persons")
            all_sd_persons_changed = self.get_sd_persons_changed(
                self.from_date, self.to_date
            )

        logger.info(f"Number of changed persons: {len(all_sd_persons_changed)}")
        all_sd_persons_changed = tqdm(all_sd_persons_changed, desc="update persons")
        real_sd_persons_changed = filter(skip_fictional_users, all_sd_persons_changed)
        sd_persons_changed = map(extract_cpr_and_name, real_sd_persons_changed)

        sd_persons_iter1, sd_persons_iter2 = tee(sd_persons_changed)
        mo_persons_iter = map(fetch_mo_person, sd_persons_iter2)

        person_pairs = zip(sd_persons_iter1, mo_persons_iter)
        has_mo_person = itemgetter(1)

        new_pairs, current_pairs = partition(has_mo_person, person_pairs)

        # Update the names of the persons already in MO
        for sd_person, mo_person in current_pairs:
            given_name = sd_person.given_name or mo_person.get("givenname", "")
            surname = sd_person.surname or mo_person.get("surname", "")
            sd_name = f"{sd_person.given_name} {sd_person.surname}"
            if mo_person["name"] == sd_name:
                continue
            uuid = mo_person["uuid"]

            upsert_employee(str(uuid), given_name, surname, sd_person.cpr)

        # Create new SD persons in MO
        for sd_person, _ in new_pairs:
            given_name = sd_person.given_name or ""
            surname = sd_person.surname or ""

            # Use previously created UUID (if such exists) for MO person
            # to be created
            forced_uuid = self.employee_forced_uuids.get(sd_person.cpr)
            sam_account_name, object_guid = self._fetch_ad_information(sd_person.cpr)

            uuid = None
            if forced_uuid:
                uuid = forced_uuid
                logger.info("Employee in force list: {}".format(uuid))
            elif object_guid:
                uuid = object_guid
                logger.debug("Using ObjectGuid as MO UUID: {}".format(uuid))
            else:
                uuid = uuid4()
                logger.debug(
                    "User not in MO, UUID list or AD, assigning UUID: {}".format(uuid)
                )

            return_uuid = upsert_employee(str(uuid), given_name, surname, sd_person.cpr)

            if sam_account_name:
                # Create an IT system for the person If the person is found in the AD
                create_itsystem_connection(sam_account_name, return_uuid)

    def _compare_dates(self, first_date, second_date, expected_diff=1):
        """
        Return true if the amount of days between second and first is smaller
        than  expected_diff.
        """
        first = datetime.datetime.strptime(first_date, "%Y-%m-%d")
        second = datetime.datetime.strptime(second_date, "%Y-%m-%d")
        delta = second - first
        # compare = first + datetime.timedelta(days=expected_diff)
        compare = abs(delta.days) <= expected_diff
        logger.debug(
            "Compare. First: {}, second: {}, expected: {}, compare: {}".format(
                first, second, expected_diff, compare
            )
        )
        return compare

    def _validity(self, engagement_info, original_end=None, cut=False):
        """
        Extract a validity object from the supplied SD information.
        If the validity extends outside the current engagment, the
        change is either refused (by returning None) or cut to the
        length of the current engagment.
        :param engagement_info: The SD object to extract from.
        :param orginal_end: The engagment end to compare with.
        :param cut: If True the returned validity will cut to fit
        rather than rejeted, if the validity is too long.
        :return: A validity dict suitable for a MO payload. None if
        the change is rejected.
        """
        from_date = engagement_info["ActivationDate"]
        to_date = engagement_info["DeactivationDate"]

        if original_end is not None:
            edit_from = datetime.datetime.strptime(from_date, "%Y-%m-%d")
            edit_end = datetime.datetime.strptime(to_date, "%Y-%m-%d")
            eng_end = datetime.datetime.strptime(original_end, "%Y-%m-%d")
            if edit_from >= eng_end:
                logger.info("This edit starts after the end of the engagement")
                return None

            if edit_end > eng_end:
                if cut:
                    to_date = datetime.datetime.strftime(eng_end, "%Y-%m-%d")
                else:
                    logger.info("This edit would have extended outside engagement")
                    return None

        if to_date == "9999-12-31":
            to_date = None
        validity = {"from": from_date, "to": to_date}

        return validity

    def _refresh_mo_engagements(self, person_uuid):
        self.mo_engagements_cache.pop(person_uuid, None)

    def _fetch_mo_engagements(self, person_uuid):
        if person_uuid in self.mo_engagements_cache:
            return self.mo_engagements_cache[person_uuid]

        mo_engagements = self.helper.read_user_engagement(
            person_uuid, read_all=True, only_primary=True, use_cache=False
        )
        self.mo_engagements_cache[person_uuid] = mo_engagements
        return mo_engagements

    def _find_engagement(self, job_id, person_uuid):
        try:
            user_key = str(int(job_id)).zfill(5)
        except ValueError:  # We will end here, if int(job_id) fails
            user_key = job_id

        logger.debug(
            "Find engagement, from date: {}, user_key: {}".format(
                self.from_date, user_key
            )
        )

        mo_engagements = self._fetch_mo_engagements(person_uuid)

        relevant_engagements = filter(
            lambda mo_eng: mo_eng["user_key"] == user_key, mo_engagements
        )
        relevant_engagement = last(relevant_engagements, None)

        if relevant_engagement is None:
            msg = "Fruitlessly searched for {} in {}".format(job_id, mo_engagements)
            logger.info(msg)
        return relevant_engagement

    def _create_class(self, payload):
        """Create a new class using the provided class payload.

        Args:
            payload: A class created using sd_payloads.* via lora_klasse

        Returns:
            uuid of the newly created class.
        """
        response = requests.post(
            url=self.settings.mox_base + "/klassifikation/klasse", json=payload
        )
        assert response.status_code == 201
        return response.json()["uuid"]

    def _create_engagement_type(self, engagement_type_ref, job_position):
        # Could not fetch, attempt to create it
        logger.warning(
            "Missing engagement_type: {} (now creating)".format(engagement_type_ref)
        )
        payload = sd_payloads.engagement_type(
            engagement_type_ref, job_position, self.org_uuid, self.engagement_type_facet
        )
        engagement_type_uuid = self._create_class(payload)
        self.engagement_types[engagement_type_ref] = engagement_type_uuid

        self.job_sync.sync_from_sd(job_position, refresh=True)

        return engagement_type_uuid

    def _create_professions(self, job_function, job_position):
        # Could not fetch, attempt to create it
        logger.warning("Missing profession: {} (now creating)".format(job_function))
        payload = sd_payloads.profession(
            job_function, self.org_uuid, self.job_function_facet
        )
        job_uuid = self._create_class(payload)
        self.job_functions[job_function] = job_uuid

        self.job_sync.sync_from_sd(job_position, refresh=True)

        return job_uuid

    def _fetch_engagement_type(self, job_position):
        """Fetch an engagement type UUID, create if missing.

        Args:
            engagement_type_ref: String of the expected engagement_type name

        Returns:
            uuid of the engagement type or None if it could not be created.
        """
        # Attempt to fetch the engagement type
        engagement_type_ref = "engagement_type" + job_position
        engagement_type_uuid = self.engagement_types.get(engagement_type_ref)
        if engagement_type_uuid:
            return engagement_type_uuid
        return self._create_engagement_type(engagement_type_ref, job_position)

    def _fetch_professions(self, job_function, job_position):
        """Fetch an job function UUID, create if missing.

        This function does not depend on self.use_jpi, as the argument is just a
        string. If self.use_jpi is true, the string will be the SD
        JobPositionIdentifier, otherwise it will be the actual job name.

        Args:
            emp_name: Overloaded job identifier string / employment name.

        Returns:
            uuid of the job function or None if it could not be created.
        """
        # Add new profssions to LoRa
        job_uuid = self.job_functions.get(job_function)
        if job_uuid:
            return job_uuid
        return self._create_professions(job_function, job_position)

    def create_leave(self, status, job_id, person_uuid: str):
        """Create a leave for a user"""
        logger.info("Create leave, job_id: {}, status: {}".format(job_id, status))
        # TODO: This code potentially creates duplicated leaves.

        # Notice, the expected and desired behaviour for leaves is for the engagement
        # to continue during the leave. It turns out this is actually what happens
        # because a leave is apparently always accompanied by a worktime-update that
        # forces an edit to the engagement that will extend it to span the
        # leave. If this ever turns out not to hold, add a dummy-edit to the
        # engagement here.
        mo_eng = self._find_engagement(job_id, person_uuid)
        payload = sd_payloads.create_leave(
            mo_eng, person_uuid, str(self.leave_uuid), job_id, self._validity(status)
        )

        response = self.helper._mo_post("details/create", payload)
        assert response.status_code == 201

    def create_association(self, department, person_uuid, job_id, validity):
        """Create a association for a user"""
        logger.info("Consider to create an association")
        associations = self.helper.read_user_association(
            person_uuid, read_all=True, only_primary=True
        )
        logger.debug("Associations read from MO: {}".format(associations))
        hit = False
        for association in associations:
            if (
                association["validity"] == validity
                and association["org_unit"]["uuid"] == department
            ):
                hit = True
        if not hit:
            logger.info("Association needs to be created")
            payload = sd_payloads.create_association(
                department, person_uuid, self.association_uuid, job_id, validity
            )
            response = self.helper._mo_post("details/create", payload)
            assert response.status_code == 201
        else:
            logger.info("No new Association is needed")

    def apply_NY_logic(self, org_unit, job_id, validity, person_uuid) -> str:
        msg = "Apply NY logic for job: {}, unit: {}, validity: {}"
        logger.debug(msg.format(job_id, org_unit, validity))
        too_deep = self.settings.sd_import_too_deep
        # Move users and make associations according to NY logic
        today = datetime.datetime.today()
        ou_info = self.helper.read_ou(org_unit, use_cache=False)
        if "status" in ou_info:
            # This unit does not exist, read its state in the not-too
            # distant future.
            fix_date = today + datetime.timedelta(weeks=80)
            self.department_fixer.fix_or_create_branch(org_unit, fix_date)
            ou_info = self.helper.read_ou(org_unit, use_cache=False)

        if ou_info["org_unit_level"]["user_key"] in too_deep:
            self.create_association(org_unit, person_uuid, job_id, validity)

        # logger.debug('OU info is currently: {}'.format(ou_info))
        while ou_info["org_unit_level"]["user_key"] in too_deep:
            ou_info = ou_info["parent"]
            logger.debug("Parent unit: {}".format(ou_info))
        org_unit = ou_info["uuid"]

        return org_unit

    def create_new_engagement(self, engagement, status, cpr, person_uuid):
        """
        Create a new engagement
        AD integration handled in check for primary engagement.
        """

        # beware - name engagement_info used for engagement in engagement_components
        user_key, engagement_info = engagement_components(engagement)
        if not engagement_info["departments"] or not engagement_info["professions"]:
            # I am looking into the possibility that creating AND finishing
            # an engagement in the past gives the problem that the engagement
            # is reported to this function without the components needed to create
            # the engagement in os2mo

            # to fix the problem we get the information for the employment at the
            # activation date

            # use a local engagement copy so we don't spill into the rest of the program
            engagement = dict(engagement)

            activation_date_info = read_employment_at(
                employment_id=engagement["EmploymentIdentifier"],
                settings=self.settings,
                effective_date=datetime.datetime.strptime(
                    status["ActivationDate"], "%Y-%m-%d"
                ).date(),
            )

            # at least check the cpr

            if cpr != activation_date_info["PersonCivilRegistrationIdentifier"]:
                logger.error(
                    "wrong cpr %r for position %r at date %r",
                    activation_date_info["PersonCivilRegistrationIdentifier"],
                    engagement["EmploymentIdentifier"],
                    status["ActivationDate"],
                )
                raise ValueError("unexpected cpr, see log")

            activation_date_engagement = activation_date_info["Employment"]
            _, activation_date_engagement_info = engagement_components(
                activation_date_engagement
            )

            # fill out the missing values
            if not engagement_info["departments"]:
                engagement_info["departments"] = activation_date_engagement_info[
                    "departments"
                ]

            if not engagement_info["professions"]:
                engagement_info["professions"] = activation_date_engagement_info[
                    "professions"
                ]

        job_position = engagement_info["professions"][0]["JobPositionIdentifier"]

        if job_position in self.skip_job_functions:
            logger.info("Skipping {} due to job_pos_id".format(engagement))
            return None

        validity = self._validity(status)
        also_edit = False
        if (
            len(engagement_info["professions"]) > 1
            or len(engagement_info["working_time"]) > 1
            or len(engagement_info["departments"]) > 1
        ):
            also_edit = True
        logger.debug("Create new engagement: also_edit: {}".format(also_edit))

        try:
            org_unit = engagement_info["departments"][0]["DepartmentUUIDIdentifier"]
            logger.info("Org unit for new engagement: {}".format(org_unit))
            org_unit = self.apply_NY_logic(org_unit, user_key, validity, person_uuid)
        except IndexError:
            msg = "No unit for engagement {}".format(user_key)
            logger.error(msg)
            raise Exception(msg)

        try:
            emp_name = engagement_info["professions"][0]["EmploymentName"]
        except (KeyError, IndexError):
            emp_name = "Ukendt"

        job_function = emp_name
        if self.use_jpi:
            job_function = job_position

        primary = self.primary_types["non_primary"]
        if status["EmploymentStatusCode"] == "0":
            primary = self.primary_types["no_salary"]

        engagement_type = self.determine_engagement_type(engagement, job_position)
        if engagement_type is None:
            return False

        extension_field = self.settings.sd_employment_field
        extension = {}
        if extension_field is not None:
            extension = {extension_field: emp_name}

        job_function_uuid = self._fetch_professions(job_function, job_position)

        payload = sd_payloads.create_engagement(
            org_unit=org_unit,
            person_uuid=person_uuid,
            job_function=job_function_uuid,
            engagement_type=engagement_type,
            primary=primary,
            user_key=user_key,
            engagement_info=engagement_info,
            validity=validity,
            **extension,
        )
        response = self.helper._mo_post("details/create", payload)
        assert response.status_code == 201

        self._refresh_mo_engagements(person_uuid)
        logger.info("Engagement {} created".format(user_key))

        if also_edit:
            # This will take of the extra entries
            self.edit_engagement(engagement, person_uuid)

        return True

    def _terminate_engagement(
        self,
        user_key: str,
        person_uuid: str,  # TODO: change type to UUID
        from_date: str,  # TODO: Introduce MO date version
        to_date: Optional[str] = None,
    ) -> bool:
        """
        Terminate an employment (engagement) in MO. Since this function calls
        MO, the parameters are adapted to accommodate MO instead of SD Løn,
        i.e. SD dates should be converted to MO dates before the function is
        invoked.

        Args:
            user_key: SD Løn employment ID. Used as BVN (user_key) in MO.
            person_uuid: The employee UUID in MO.
            from_date: The MO "from" date (to be set in virkning).
            to_date: The MO "to" date (to be set in virkning).

        Returns:
            `True` if the termination in MO was successful and `False`
            otherwise
        """
        mo_engagement = self._find_engagement(user_key, person_uuid)

        if not mo_engagement:
            logger.warning(f"Terminating non-existing job: {user_key}!")
            return False

        # TODO: remove this hack when MO is updated to a version greater
        #  than 2.7.0 !!

        validity = {"from": from_date, "to": to_date}
        if self.terminate_engagement_with_to_only:
            validity = {"to": sd_to_mo_termination_date(from_date)}

        # TODO: use/create termination object from RA Models
        payload = {
            "type": "engagement",
            "uuid": mo_engagement["uuid"],
            "validity": validity,
        }

        logger.debug("Terminate payload: {}".format(payload))
        response = self.helper._mo_post("details/terminate", payload)
        logger.debug("Terminate response: {}".format(response.text))
        mora_assert(response)

        self._refresh_mo_engagements(person_uuid)

        return True

    def edit_engagement_department(self, engagement, mo_eng, person_uuid):
        job_id, engagement_info = engagement_components(engagement)
        for department in engagement_info["departments"]:
            logger.info("Change department of engagement {}:".format(job_id))
            logger.debug("Department object: {}".format(department))

            validity = self._validity(department, mo_eng["validity"]["to"], cut=True)
            if validity is None:
                continue

            logger.debug("Validity of this department change: {}".format(validity))
            org_unit = department["DepartmentUUIDIdentifier"]
            if org_unit is None:
                logger.warning(
                    "DepartmentUUIDIdentifier was None, attempting GetDepartment"
                )
                # This code should not be necessary, but SD returns bad data.
                # Sometimes the UUID is missing, even if it can be looked up?
                url = "GetDepartment20111201"
                params = {
                    "ActivationDate": self.from_date.strftime("%d.%m.%Y"),
                    "DeactivationDate": self.from_date.strftime("%d.%m.%Y"),
                    "DepartmentNameIndicator": "true",
                    "UUIDIndicator": "true",
                    "DepartmentIdentifier": department["DepartmentIdentifier"],
                }
                response = sd_lookup(url, settings=self.settings, params=params)
                logger.warning("GetDepartment returned: {}".format(response))
                org_unit = response["Department"]["DepartmentUUIDIdentifier"]
                if org_unit is None:
                    logger.fatal("DepartmentUUIDIdentifier was None inside failover.")
                    sys.exit(1)

            associations = self.helper.read_user_association(person_uuid, read_all=True)
            logger.debug("User associations: {}".format(associations))
            current_association = None
            # TODO: This is a filter + next (only?)
            for association in associations:
                if association["user_key"] == job_id:
                    current_association = association["uuid"]

            if current_association:
                logger.debug("We need to move {}".format(current_association))
                data = {"org_unit": {"uuid": org_unit}, "validity": validity}
                payload = sd_payloads.association(data, current_association)
                logger.debug("Association edit payload: {}".format(payload))
                response = self.helper._mo_post("details/edit", payload)
                mora_assert(response)

            org_unit = self.apply_NY_logic(org_unit, job_id, validity, person_uuid)

            logger.debug("New org unit for edited engagement: {}".format(org_unit))
            data = {"org_unit": {"uuid": org_unit}, "validity": validity}
            payload = sd_payloads.engagement(data, mo_eng)
            response = self.helper._mo_post("details/edit", payload)
            mora_assert(response)

    def determine_engagement_type(self, engagement, job_position):
        split = self.settings.sd_monthly_hourly_divide
        employment_id = calc_employment_id(engagement)
        if employment_id["value"] < split:
            return self.engagement_types.get("månedsløn")
        # XXX: Is the first condition not implied by not hitting the above case?
        if (split - 1) < employment_id["value"] < 999999:
            return self.engagement_types.get("timeløn")
        # This happens if EmploymentID is not a number
        # XXX: Why are we checking against 999999 instead of checking the type?
        # Once we get here, we know that it is a no-salary employee

        # We should not create engagements (or engagement_types) for engagements
        # with too low of a job_position id compared to no_salary_minimum_id.
        if (
            self.no_salary_minimum is not None
            and int(job_position) < self.no_salary_minimum
        ):
            message = "No salary employee, with too low job_position id"
            logger.warning(message)
            return None

        # We need a special engagement type for the engagement.
        # We will try to fetch and try to create it if we cannot find it.
        logger.info("Non-nummeric id. Job pos id: {}".format(job_position))
        return self._fetch_engagement_type(job_position)

    def edit_engagement_type(self, engagement, mo_eng):
        job_id, engagement_info = engagement_components(engagement)
        for profession_info in engagement_info["professions"]:
            logger.info("Change engagement type of engagement {}".format(job_id))
            job_position = profession_info["JobPositionIdentifier"]

            validity = self._validity(
                profession_info, mo_eng["validity"]["to"], cut=True
            )
            if validity is None:
                continue

            engagement_type = self.determine_engagement_type(engagement, job_position)
            if engagement_type is None:
                continue
            data = {"engagement_type": {"uuid": engagement_type}, "validity": validity}
            payload = sd_payloads.engagement(data, mo_eng)
            logger.debug("Update engagement type payload: {}".format(payload))
            response = self.helper._mo_post("details/edit", payload)
            mora_assert(response)

    def edit_engagement_profession(self, engagement, mo_eng):
        job_id, engagement_info = engagement_components(engagement)
        for profession_info in engagement_info["professions"]:
            logger.info("Change profession of engagement {}".format(job_id))
            job_position = profession_info["JobPositionIdentifier"]

            # The variability handling introduced in the following lines
            # (based on the value of job_position) is not optimal, i.e.
            # a parametric if-switch is used, where a strategy pattern would
            # be more appropriate. However, due to all the hard couplings in
            # the code, a strategy pattern is not feasible for now. Let's
            # leave it as is until the whole SD code base is rewritten

            if not is_employment_id_and_no_salary_minimum_consistent(
                engagement, self.no_salary_minimum
            ):
                sd_from_date = profession_info["ActivationDate"]
                sd_to_date = profession_info["DeactivationDate"]
                self._terminate_engagement(
                    mo_eng["user_key"],
                    mo_eng["person"]["uuid"],
                    sd_from_date,
                    sd_to_mo_termination_date(sd_to_date),
                )
            else:
                emp_name = profession_info.get("EmploymentName", job_position)
                validity = self._validity(
                    profession_info, mo_eng["validity"]["to"], cut=True
                )
                if validity is None:
                    continue

                job_function = emp_name
                if self.use_jpi:
                    job_function = job_position
                logger.debug("Employment name: {}".format(job_function))

                ext_field = self.settings.sd_employment_field
                extention = {}
                if ext_field is not None:
                    extention = {ext_field: emp_name}

                job_function_uuid = self._fetch_professions(job_function, job_position)

                data = {
                    "job_function": {"uuid": job_function_uuid},
                    "validity": validity,
                }
                data.update(extention)
                payload = sd_payloads.engagement(data, mo_eng)
                logger.debug("Update profession payload: {}".format(payload))

                response = self.helper._mo_post("details/edit", payload)
                mora_assert(response)

    def edit_engagement_worktime(self, engagement, mo_eng):
        job_id, engagement_info = engagement_components(engagement)
        for worktime_info in engagement_info["working_time"]:
            logger.info("Change working time of engagement {}".format(job_id))

            validity = self._validity(worktime_info, mo_eng["validity"]["to"], cut=True)
            if validity is None:
                continue

            working_time = float(worktime_info["OccupationRate"])
            data = {"fraction": int(working_time * 1000000), "validity": validity}
            payload = sd_payloads.engagement(data, mo_eng)
            logger.debug("Change worktime, payload: {}".format(payload))
            response = self.helper._mo_post("details/edit", payload)
            mora_assert(response)

    def _set_non_primary(self, status, mo_eng):
        logger.debug("Setting non-primary for: {}".format(mo_eng["uuid"]))

        validity = self._validity(status)
        logger.debug("Validity for edit: {}".format(validity))

        data = {
            "primary": {"uuid": self.primary_types["non_primary"]},
            "validity": validity,
        }
        payload = sd_payloads.engagement(data, mo_eng)
        logger.debug("Setting non-primary payload: {}".format(payload))

        response = self.helper._mo_post("details/edit", payload)
        mora_assert(response)

    def edit_engagement(self, engagement, person_uuid):
        """
        Edit an engagement
        """
        employment_id, engagement_info = engagement_components(engagement)
        mo_eng = self._find_engagement(employment_id, person_uuid)

        employment_consistent = is_employment_id_and_no_salary_minimum_consistent(
            engagement, self.no_salary_minimum
        )

        if mo_eng is None:
            if employment_consistent:
                create_engagement(self, employment_id, person_uuid)
            return

        update_existing_engagement(self, mo_eng, engagement, person_uuid)

    def _handle_employment_status_changes(
        self, cpr: str, sd_employment: OrderedDict, person_uuid: str
    ) -> bool:
        """
        Update MO with SD employment changes.

        Args:
            cpr: The CPR number of the person.
            sd_employment: The SD employment (see example below).
            person_uuid: The UUID of the MO employee.

        Returns:
            `True` if an employment CRUD operation has been executed and
            `False` otherwise.

        Examples:
            The sd_employment could for example look like this:
                ```python
                OrderedDict([
                    ('EmploymentIdentifier', '12345'),
                    ('EmploymentDate', '2020-11-10'),
                    ('EmploymentDepartment', OrderedDict([
                        ('@changedAtDate', '2020-11-10'),
                        ('ActivationDate', '2020-11-10'),
                        ('ActivationTime', '06:00'),
                        ('DeactivationDate', '9999-12-31'),
                        ('DepartmentIdentifier', 'department_id'),
                        ('DepartmentUUIDIdentifier', 'department_uuid')
                    ])),
                    ('Profession', OrderedDict([
                        ('@changedAtDate', '2020-11-10'),
                        ('ActivationDate', '2020-11-10'),
                        ('ActivationTime', '06:00'),
                        ('DeactivationDate', '9999-12-31'),
                        ('JobPositionIdentifier', '1'),
                        ('EmploymentName', 'chief'),
                        ('AppointmentCode', '0')
                    ])),
                    ('EmploymentStatus', [
                        OrderedDict([
                            ('@changedAtDate', '2020-11-10'),
                            ('ActivationDate', '2020-11-10'),
                            ('ActivationTime', '06:00'),
                            ('DeactivationDate', '2021-02-09'),
                            ('EmploymentStatusCode', '1')
                        ]),
                        OrderedDict([
                            ('@changedAtDate', '2020-11-10'),
                            ('ActivationDate', '2021-02-10'),
                            ('ActivationTime', '06:00'),
                            ('DeactivationDate', '9999-12-31'),
                            ('EmploymentStatusCode', '8')
                        ])
                    ])
                ])
                ```
        """

        skip = False
        # The EmploymentStatusCode can take a number of magical values.
        # that must be handled separately.
        employment_id, eng = engagement_components(sd_employment)
        for status in eng["status_list"]:
            logger.info("Status is: {}".format(status))
            code = status["EmploymentStatusCode"]
            code = EmploymentStatus(code)

            if code == EmploymentStatus.AnsatUdenLoen:
                logger.info("Status 0. Cpr: {}, job: {}".format(cpr, employment_id))
                mo_eng = self._find_engagement(employment_id, person_uuid)
                if mo_eng:
                    logger.info("Status 0, edit eng {}".format(mo_eng["uuid"]))

                    self._set_non_primary(status, mo_eng)

                    self.edit_engagement(sd_employment, person_uuid)
                else:
                    logger.info("Status 0, create new engagement")
                    if is_employment_id_and_no_salary_minimum_consistent(
                        sd_employment, self.no_salary_minimum
                    ):
                        self.create_new_engagement(
                            sd_employment, status, cpr, person_uuid
                        )
                skip = True
            elif code == EmploymentStatus.AnsatMedLoen:
                logger.info("Setting {} to status 1".format(employment_id))
                mo_eng = self._find_engagement(employment_id, person_uuid)
                if mo_eng:
                    logger.info("Status 1, edit eng. {}".format(mo_eng["uuid"]))

                    self._set_non_primary(status, mo_eng)
                    self._refresh_mo_engagements(person_uuid)

                    self.edit_engagement(sd_employment, person_uuid)
                else:
                    logger.info("Status 1: Create new engagement")
                    logger.debug(sd_employment)
                    if is_employment_id_and_no_salary_minimum_consistent(
                        sd_employment, self.no_salary_minimum
                    ):
                        self.create_new_engagement(
                            sd_employment, status, cpr, person_uuid
                        )
                skip = True
            elif code == EmploymentStatus.Orlov:
                mo_eng = self._find_engagement(employment_id, person_uuid)
                if not mo_eng:
                    logger.info("Leave for non existent eng., create one")
                    if is_employment_id_and_no_salary_minimum_consistent(
                        sd_employment, self.no_salary_minimum
                    ):
                        self.create_new_engagement(
                            sd_employment, status, cpr, person_uuid
                        )
                logger.info("Create a leave for {} ".format(cpr))
                self.create_leave(status, employment_id, person_uuid)
            elif code in EmploymentStatus.let_go():
                sd_from_date = status["ActivationDate"]
                sd_to_date = status["DeactivationDate"]
                logger.info("Terminate {}, job_id {} ".format(cpr, employment_id))
                success = self._terminate_engagement(
                    user_key=employment_id,
                    person_uuid=person_uuid,
                    from_date=sd_from_date,
                    to_date=sd_to_mo_termination_date(sd_to_date),
                )
                if not success:
                    logger.error("Problem with job-id: {}".format(employment_id))
                    skip = True
            elif code == EmploymentStatus.Slettet:

                # TODO: rename user_key to something unique in MO when employee
                # is terminated.
                #
                # The reason for this is that SD Løn sometimes reuses the same
                # job_id for *different* persons, e.g. if a person with
                # job_id=12345 is set to "Slettet" then a different newly
                # employed SD person can get the SAME job_id!
                #
                # In MO we therefore have to do the following. When a MO person
                # is terminated, we have to make sure the the user_key (BVN) of
                # that user is changed to some unique, e.g. "old user_key +
                # UUID (or date)". In that way we can avoid user_key conflicts
                # between different employees.
                #
                # Note that an SD person can jump from any status to "Slettet"

                for mo_eng in self._fetch_mo_engagements(person_uuid):
                    if mo_eng["user_key"] == employment_id:
                        sd_from_date = status["ActivationDate"]
                        logger.info("Status S: Terminate {}".format(employment_id))
                        self._terminate_engagement(
                            user_key=employment_id,
                            person_uuid=person_uuid,
                            from_date=sd_from_date,
                        )
                skip = True
        return skip

    def _update_user_employments(
        self, cpr: str, sd_employments, person_uuid: str
    ) -> None:
        for sd_employment in sd_employments:
            job_id, eng = engagement_components(sd_employment)
            logger.info("Update Job id: {}".format(job_id))
            logger.debug("SD Engagement: {}".format(sd_employment))
            # If status is present, we have a potential creation
            if eng["status_list"] and self._handle_employment_status_changes(
                cpr, sd_employment, person_uuid
            ):
                continue
            self.edit_engagement(sd_employment, person_uuid)

    def update_all_employments(
        self, in_cpr: Optional[str] = None, dry_run: bool = False
    ) -> None:
        if in_cpr is not None:
            employments_changed = self.read_employment_changed(in_cpr=in_cpr)
        else:
            logger.info("Update all employments")
            employments_changed = self.read_employment_changed()

        logger.info("Update a total of {} employments".format(len(employments_changed)))

        employments_changed = tqdm(employments_changed, desc="update employments")
        employments_changed = filter(skip_fictional_users, employments_changed)
        recalculate_users: Set[UUID] = set()

        for employment in employments_changed:
            cpr = employment["PersonCivilRegistrationIdentifier"]
            sd_employments = ensure_list(employment["Employment"])

            logger.info("---------------------")
            logger.info("We are now updating {}".format(cpr))
            logger.debug("From date: {}".format(self.from_date))
            logger.debug("To date: {}".format(self.to_date))
            logger.debug("Employment: {}".format(employment))

            mo_person = self.helper.read_user(user_cpr=cpr, org_uuid=self.org_uuid)
            # Person not in MO, but they should be
            if not mo_person:
                logger.warning("This person should be in MO, but is not")
                try:
                    self.update_changed_persons(in_cpr=cpr, dry_run=dry_run)
                    mo_person = self.helper.read_user(
                        user_cpr=cpr, org_uuid=self.org_uuid
                    )
                except Exception as exp:
                    logger.error("Unable to find person in MO, SD error: " + str(exp))
                    continue

            person_uuid = mo_person["uuid"]

            self._refresh_mo_engagements(person_uuid)
            if dry_run:
                print("Dry-run: update_user_employments", sd_employments, person_uuid)
            else:
                self._update_user_employments(cpr, sd_employments, person_uuid)
            # Re-calculate primary after all updates for user has been performed.
            recalculate_users.add(person_uuid)

        logger.info("Beginning recalculation of all users...")

        for user_uuid in recalculate_users:
            if dry_run:
                print("Dry-run: recalculate_user", user_uuid)
                continue

            try:
                self.updater.recalculate_user(user_uuid)
            except NoPrimaryFound:
                logger.warning("Could not find primary for: {}".format(user_uuid))


def _local_db_insert(insert_tuple):
    settings = load_settings()
    conn = sqlite3.connect(
        settings["integrations.SD_Lon.import.run_db"],
        detect_types=sqlite3.PARSE_DECLTYPES,
    )
    c = conn.cursor()
    query = "INSERT INTO runs (from_date, to_date, status) VALUES (?, ?, ?)"
    final_tuple = (
        insert_tuple[0],
        insert_tuple[1],
        insert_tuple[2].format(datetime.datetime.now()),
    )
    c.execute(query, final_tuple)
    conn.commit()
    conn.close()


def initialize_changed_at(from_date, run_db, force=False):
    if not run_db.is_file():
        raise Exception("RunDB not created, use 'db_overview.py' to create it")

    _local_db_insert((from_date, from_date, "Running since {}"))
    settings = get_changed_at_settings()

    logger.info("Start initial ChangedAt")
    sd_updater = ChangeAtSD(settings, from_date)
    sd_updater.update_changed_persons()
    sd_updater.update_all_employments()
    logger.info("Ended initial ChangedAt")

    _local_db_insert((from_date, from_date, "Initial import: {}"))


def get_from_date(run_db, force: bool = False) -> datetime.datetime:
    db_overview = DBOverview(run_db)
    # To date from last entries, becomes from_date for current entry
    from_date, status = cast(
        Tuple[datetime.datetime, str], db_overview._read_last_line("to_date", "status")
    )
    if "Running" in status:
        if force:
            db_overview.delete_last_row()
            from_date, status = cast(
                Tuple[datetime.datetime, str],
                db_overview._read_last_line("to_date", "status"),
            )
        else:
            logging.error("Previous ChangedAt run did not return!")
            raise click.ClickException("Previous ChangedAt run did not return!")
    return from_date


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--init",
    is_flag=True,
    type=click.BOOL,
    default=False,
    help="Initialize a new rundb",
)
@click.option(
    "--force",
    is_flag=True,
    type=click.BOOL,
    default=False,
    help="Ignore previously unfinished runs",
)
@click.option(
    "--from-date",
    type=click.DateTime(),
    help="Global import from-date, only used if init is True",
)
def changed_at(init: bool, force: bool, from_date: datetime.datetime):
    """Tool to delta synchronize with MO with SD."""
    setup_logging()

    settings = get_changed_at_settings()
    run_db = settings.sd_import_run_db

    logger.info("***************")
    logger.info("Program started")

    if init:
        if not from_date:
            from_date = date_to_datetime(settings.sd_global_from_date)
        run_db_path = pathlib.Path(run_db)

        initialize_changed_at(from_date, run_db_path, force=True)
        exit()

    from_date = get_from_date(run_db, force=force)
    to_date = datetime.datetime.now()
    dates = gen_date_intervals(from_date, to_date)
    for from_date, to_date in dates:
        logger.info("Importing {} to {}".format(from_date, to_date))
        _local_db_insert((from_date, to_date, "Running since {}"))

        logger.info("Start ChangedAt module")
        sd_updater = ChangeAtSD(settings, from_date, to_date)

        logger.info("Update changed persons")
        sd_updater.update_changed_persons()

        logger.info("Update all employments")
        sd_updater.update_all_employments()

        _local_db_insert((from_date, to_date, "Update finished: {}"))

        logger.info("Program stopped.")


@cli.command()
@click.option(
    "--cpr",
    required=True,
    type=click.STRING,
    help="CPR number of the person to import",
)
@click.option(
    "--from-date",
    type=click.DateTime(),
    help="Global import from-date",
)
@click.option(
    "--dry-run", is_flag=True, default=False, help="Dry-run making no actual changes."
)
def import_single_user(cpr: str, from_date: datetime.datetime, dry_run: bool):
    """Import a single user into MO."""

    settings = get_changed_at_settings()
    if not from_date:
        from_date = date_to_datetime(settings.sd_global_from_date)

    sd_updater = ChangeAtSD(settings, from_date, None)
    sd_updater.update_changed_persons(cpr, dry_run=dry_run)
    sd_updater.update_all_employments(cpr, dry_run=dry_run)


@cli.command()
@click.option(
    "--from-date",
    type=click.DateTime(),
    help="Global import from-date",
)
@click.option(
    "--dry-run", is_flag=True, default=False, help="Dry-run making no actual changes."
)
def import_state(from_date: datetime.datetime, dry_run: bool):
    """Import engagement changes for all users."""
    setup_logging()
    settings = get_changed_at_settings()

    if not from_date:
        from_date = date_to_datetime(settings.sd_global_from_date)

    sd_updater = ChangeAtSD(settings, from_date, None)
    sd_updater.update_changed_persons(dry_run=dry_run)
    sd_updater.update_all_employments(dry_run=dry_run)


if __name__ == "__main__":
    cli()



import unittest
from unittest.mock import patch

from helpers import dummy_settings
from parameterized import parameterized

from integrations.os2sync import os2mo
from integrations.os2sync.os2mo import get_sts_user as os2mo_get_sts_user
from integrations.os2sync.tests.helpers import MoEmployeeMixin
from integrations.os2sync.tests.helpers import NICKNAME_TEMPLATE


class TestGetStsUser(unittest.TestCase, MoEmployeeMixin):
    maxDiff = None

    def setUp(self):
        super().setUp()
        self._uuid = "mock-uuid"
        self._user_key = "mock-user-key"

    @parameterized.expand(
        [
            # Test without template
            (
                None,  # template config
                dict(nickname=False),  # mo employee response kwargs
                "name",  # key of expected value for `Name`
            ),
            # Test with template: user has no nickname
            (
                {"person.name": NICKNAME_TEMPLATE},  # template config
                dict(nickname=False),  # mo employee response kwargs
                "name",  # key of expected value for `Name`
            ),
            # Test with template: user has a nickname
            (
                {"person.name": NICKNAME_TEMPLATE},  # template config
                dict(nickname=True),  # mo employee response kwargs
                "nickname",  # key of expected value for `Name`
            ),
        ]
    )
    def test_person_template_nickname(
        self,
        os2sync_templates,
        response_kwargs,
        expected_key,
    ):
        mo_employee_response = self.mock_employee_response(**response_kwargs)
        sts_user = self._run(
            mo_employee_response,
            ad_user_key=self._user_key,
            os2sync_templates=os2sync_templates,
        )
        self.assertDictEqual(
            sts_user,
            {
                "Uuid": self._uuid,
                "UserId": self._user_key,
                "Positions": [],
                "Person": {
                    "Name": mo_employee_response.json()[expected_key],
                    "Cpr": mo_employee_response.json()["cpr_no"],
                },
            },
        )

    @parameterized.expand(
        [
            # Test without an AD BVN and without template
            (
                None,  # template config
                None,  # return value of `try_get_ad_user_key`
                "mock-uuid",  # expected value of `UserId` (MO UUID)
            ),
            # Test without an AD BVN, and template which uses `user_key`
            (
                {"person.user_id": "{{ user_key }}"},  # template config
                None,  # return value of `try_get_ad_user_key`
                "testtestesen",  # expected value of `UserId` (MO BVN)
            ),
            # Test without an AD BVN, and template which uses `uuid`
            (
                {"person.user_id": "{{ uuid }}"},  # template config
                None,  # return value of `try_get_ad_user_key`
                "mock-uuid",  # expected value of `UserId` (MO UUID)
            ),
            # Test with an AD BVN, but without template
            (
                None,  # template config
                "mock-ad-bvn",  # return value of `try_get_ad_user_key`
                "mock-ad-bvn",  # expected value of `UserId` (AD BVN)
            ),
            # Test with an AD BVN, and template which uses `user_key`
            (
                {"person.user_id": "{{ user_key }}"},  # template config
                "mock-ad-bvn",  # return value of `try_get_ad_user_key`
                "mock-ad-bvn",  # expected value of `UserId` (AD BVN)
            ),
            # Test with an AD BVN, and template which uses `uuid`
            (
                {"person.user_id": "{{ uuid }}"},  # template config
                "mock-ad-bvn",  # return value of `try_get_ad_user_key`
                "mock-ad-bvn",  # expected value of `UserId` (AD BVN)
            ),
        ]
    )
    def test_user_template_user_id(
        self,
        os2sync_templates,
        given_ad_user_key,
        expected_user_id,
    ):
        mo_employee_response = self.mock_employee_response()
        sts_user = self._run(
            mo_employee_response,
            ad_user_key=given_ad_user_key,
            os2sync_templates=os2sync_templates,
        )
        self.assertDictEqual(
            sts_user,
            {
                "Uuid": self._uuid,
                "UserId": expected_user_id,
                "Positions": [],
                "Person": {
                    "Name": mo_employee_response.json()["name"],
                    "Cpr": mo_employee_response.json()["cpr_no"],
                },
            },
        )

    @patch.object(os2mo, "engagements_to_user", return_value=[])
    @patch.object(os2mo, "addresses_to_user", return_value=[])
    def _run(
        self,
        response,
        engagemnts_mock,
        address_mock,
        ad_user_key=None,
        os2sync_templates=None,
    ):
        settings = dummy_settings
        settings.os2sync_xfer_cpr = True
        settings.os2sync_templates = os2sync_templates or {}
        with self._patch("os2mo_get", response):
            with self._patch("try_get_ad_user_key", ad_user_key):
                return os2mo_get_sts_user(self._uuid, [], settings=settings)

    def _patch(self, name, return_value):
        return patch.object(os2mo, name, return_value=return_value)



from unittest.mock import patch
from uuid import uuid4

import pytest
from parameterized import parameterized
from pydantic import ValidationError

from integrations.os2sync.config import get_os2sync_settings

env_municipality = "1234"
env_uuid = uuid4()
file_municipality = "5678"
file_uuid = uuid4()
file_api_url = "http://defined.in/file"


@pytest.fixture
def mock_env(monkeypatch):
    monkeypatch.setenv("MUNICIPALITY", env_municipality)
    monkeypatch.setenv("OS2SYNC_TOP_UNIT_UUID", str(env_uuid))


class TestConfig:
    dummy_config = {
        "municipality.cvr": file_municipality,
        "os2sync_top_unit_uuid": file_uuid,
        "os2sync.api_url": file_api_url,
    }

    @patch("integrations.os2sync.config.load_settings", return_value={})
    def test_no_settings(self, settings_mock):
        with pytest.raises(ValidationError):
            get_os2sync_settings.cache_clear()
            get_os2sync_settings()

    @patch("integrations.os2sync.config.load_settings", return_value=dummy_config)
    def test_minimal_settings_file(self, settings_mock):
        get_os2sync_settings.cache_clear()
        settings = get_os2sync_settings()
        assert settings.municipality == self.dummy_config["municipality.cvr"]
        assert (
            settings.os2sync_top_unit_uuid == self.dummy_config["os2sync_top_unit_uuid"]
        )
        assert settings.os2sync_api_url == self.dummy_config["os2sync.api_url"]

    @patch("integrations.os2sync.config.load_settings", return_value={})
    def test_minimal_settings_env(self, mock_settings_file, mock_env):
        get_os2sync_settings.cache_clear()
        settings = get_os2sync_settings()
        assert settings.municipality == env_municipality
        assert settings.os2sync_top_unit_uuid == env_uuid

    @patch("integrations.os2sync.config.load_settings", return_value=dummy_config)
    def test_env_overrides(self, mock_settings_file, mock_env):
        get_os2sync_settings.cache_clear()
        settings = get_os2sync_settings()
        # Exists in both env and file, using env
        assert settings.municipality == env_municipality
        assert settings.os2sync_top_unit_uuid == env_uuid
        # Exists only in file
        assert settings.os2sync_api_url == file_api_url

    def test_full_config(self):
        conf = {
            "municipality.cvr": "test",
            "mora.base": "http://testos2mo.dk",
            "os2sync.top_unit_uuid": uuid4(),
            "os2sync.api_url": "http://testos2sync.dk",
            "os2sync.use_lc_db": False,
            "os2sync.log_level": 1,
            "os2sync.log_file": "test.log",
            "os2sync.hash_cache": "test",
            "os2sync.xfer_cpr": True,
            "os2sync.autowash": False,
            "os2sync.ca_verify_os2sync": True,
            "os2sync.ca_verify_os2mo": True,
            "os2sync.phone_scope_classes": [uuid4(), uuid4()],
            "os2sync.email_scope_classes": [uuid4(), uuid4()],
            "os2sync.ignored.unit_levels": [uuid4(), uuid4()],
            "os2sync.ignored.unit_types": [uuid4(), uuid4()],
            "os2sync.templates": {"template": "test"},
            "os2sync.use_contact_for_tasks": False,
        }
        with patch("integrations.os2sync.config.load_settings", return_value=conf):
            get_os2sync_settings.cache_clear()
            assert get_os2sync_settings()

    # Test that wrong value types raises a validation error
    @parameterized.expand(
        [
            ({"os2sync.api_url": "Not a URL"},),
            ({"mora.base": "Not a URL"},),
            ({"os2sync.use_lc_db": "No"},),
        ]
    )
    @patch("integrations.os2sync.config.load_settings", return_value=dummy_config)
    def test_invalid_values(self, wrong_type_config, settings_mock):
        settings_mock.return_value.update(wrong_type_config)
        with pytest.raises(ValidationError):
            get_os2sync_settings.cache_clear()
            get_os2sync_settings()



#!/usr/bin/env python3
#
# Copyright (c) 2017-2018, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import os
import pickle
import logging
import requests
import xmltodict
import collections
from integrations import dawa_helper
from datetime import datetime
from uuid import UUID

LOG_LEVEL = logging.DEBUG
LOG_FILE = 'mo_integrations.log'

logger = logging.getLogger("aposImport")

for name in logging.root.manager.loggerDict:
    if name in ('aposImport', 'moImporterMoraTypes', 'moImporterMoxTypes',
                'moImporterUtilities', 'moImporterHelpers', 'mora-helper'):
        logging.getLogger(name).setLevel(LOG_LEVEL)
    else:
        logging.getLogger(name).setLevel(logging.WARNING)

logging.basicConfig(
    format='%(levelname)s %(asctime)s %(name)s %(message)s',
    level=LOG_LEVEL,
    filename=LOG_FILE
)

logger.info('Apos import started')

MUNICIPALTY_NAME = os.environ.get('MUNICIPALITY_NAME', 'APOS Import')
BASE_APOS_URL = os.environ.get('BASE_APOS_URL', 'http://localhost:8080/apos2-')

# Phone names will become a list once we have more examples of naming of
# the concept phone numbers
PHONE_NAMES = os.environ.get('PHONE_NAMES', 'Telefon').split(':')

ANSAT_UUID = os.environ.get('ANSAT_UUID', '00000000-0000-0000-0000-000000000000')
CREATE_UDVALGS_CLASSES = os.environ.get('CREATE_UDVALGS_CLASSES', 'no') == 'yes'
EMAIL_NAME = os.environ.get('EMAIL_NAME', 'Email')
MAIN_PHONE_NAME = os.environ.get('MAIN_PHONE_NAME', 'Telefon')
ALT_PHONE_NAME = os.environ.get('ALT_PHONE_NAME', None)


def _format_time(gyldighed):
    from_time = '1920-01-01'
    to_time = None
    if not gyldighed['@fra'] == '-INFINITY':
        from_time = datetime.strptime(gyldighed['@fra'], '%d/%m/%Y')
        from_time = from_time.strftime('%Y-%m-%d')
    if not gyldighed['@til'] == 'INFINITY':
        to_time = datetime.strptime(gyldighed['@til'], '%d/%m/%Y')
        to_time = to_time.strftime('%Y-%m-%d')

    if from_time is None and to_time is None:
        raise Exception('Invalid validity: {}'.format(gyldighed))
    return from_time, to_time


class AposImport(object):

    def __init__(self, importer, org_name, municipality_code, org_uuid=None, ean={}):
        self.base_url = BASE_APOS_URL

        self.importer = importer
        self.importer.add_organisation(
            uuid=org_uuid,
            identifier=org_name,
            user_key=org_name,
            municipality_code=municipality_code
        )

        self.ean = ean
        self.object_to_uuid = {}  # Mapping of Apos object ID to Apos UUID
        self.address_challenges = {}  # Needs support in dawa helper
        self.duplicate_persons = {}
        self.address_errors = {}  # Needs support in dawa helper
        self.klassifikation_errors = {}

    def _apos_lookup(self, url):
        path_url = url.replace('/', '_')
        try:
            with open('tmp/' + path_url + '.p', 'rb') as f:
                response = pickle.load(f)
        except FileNotFoundError:
            response = requests.get(self.base_url + url)
            with open('tmp/' + path_url + '.p', 'wb') as f:
                pickle.dump(response, f, pickle.HIGHEST_PROTOCOL)

        xml_response = xmltodict.parse(response.text)
        # Response is always hidden one level down
        outer_key = list(xml_response.keys())[0]
        return xml_response[outer_key]

    def read_locations(self, unit):
        # url = 'app-organisation/GetOrganisationEnhedIntegration?uuid={}'
        # integration_values = self._apos_lookup(url.format(unit['@uuid']))
        # print(integration_values)

        url = 'app-organisation/GetLocations?uuid={}'
        locations = self._apos_lookup(url.format(unit['@uuid']))

        mo_locations = []
        if int(locations['total']) == 0:
            # Return imidiately if no locations are found
            return mo_locations
        locations = locations['location']
        if not isinstance(locations, list):
            locations = [locations]

        for location in locations:
            address_string = None
            if '@adresse' not in location:
                if len(location['@navn']) > 0:
                    address = location['@navn'].split(' ')
                    if len(address) == 1:
                        continue
                    zip_code = address[-2][:4]
                    address_string = ''
                    for i in range(0, len(address) - 2):
                        address_string += address[i] + ' '
                    address_string = address_string[:-2]
            else:
                uuid = location['@adresse']
                url = 'app-part/GetAdresseList?uuid={}'
                apos_address = self._apos_lookup(url.format(uuid))
                address_string = apos_address['adresse']['@vejadresseringsnavn']
                zip_code = apos_address['adresse']['@postnummer']

            if address_string:
                dawa_uuid = dawa_helper.dawa_lookup(address_string, zip_code)
            else:
                dawa_uuid = None

            pnummer = location.get('@pnummer', None)
            primary = (location.get('@primary', None) == 'JA')
            mo_location = {'pnummer': pnummer,
                           'primary': primary,
                           'dawa_uuid': dawa_uuid}
            mo_locations.append(mo_location)
        return mo_locations

    def _read_apos_facetter(self, uuid):
        url = "app-klassifikation/GetFacetterForKlassifikation?uuid={}"
        r = self._apos_lookup(url.format(uuid))
        facetter = r['facet']
        if not isinstance(facetter, list):
            facetter = [facetter]
        facet_dict = {}
        for facet in facetter:
            beskrivelse = facet['@brugervendtnoegle']
            facet_dict[beskrivelse] = facet
        return facet_dict

    def _read_apos_klasser(self, facet_uuid):
        url = "app-klassifikation/GetKlasseForFacet?uuid={}"
        r = self._apos_lookup(url.format(facet_uuid))
        return r['klasse']

    def create_typer(self, klassifikation_uuid, facet_typer):
        """ Read all klasser from a klassifikation from Apos
        and create them in LoRa. Apos has only a single facet
        for these types, so the Apos Klassifikation will be mapped
        as a LoRa facet.
        :param klassifikation_uuid: The apos uuid for the klassifikation
        :param facet_type: Dict with MO facets as keys and APOS facets as
        values
        """
        apos_facetter = self._read_apos_facetter(klassifikation_uuid)
        for mo_facet_navn, apos_facet_navne in facet_typer.items():
            for apos_facet_navn in apos_facet_navne:
                facet = apos_facetter[apos_facet_navn]
                klasser = self._read_apos_klasser(facet['@uuid'])
                if not isinstance(klasser, list):
                    klasser = [klasser]
                for klasse in klasser:
                    logger.debug('{}: {}'.format(klasse['@uuid'], klasse['@title']))
                    if klasse['@uuid'] in PHONE_NAMES:
                        scope = 'PHONE'
                    elif klasse['@uuid'] == EMAIL_NAME:
                        scope = 'EMAIL'
                    else:
                        scope = 'TEXT'
                    try:
                        UUID(klasse['@uuid'], version=4)
                        uuid = klasse['@uuid']
                    except ValueError:
                        uuid = None

                    self.importer.add_klasse(identifier=klasse['@uuid'],
                                             uuid=uuid,
                                             title=klasse['@title'],
                                             user_key=klasse['@title'],
                                             scope=scope,
                                             facet_type_ref=mo_facet_navn)

    def create_facetter_and_klasser(self):
        url = "app-klassifikation/GetKlassifikationList"
        r = self._apos_lookup(url)
        klassifikationer = r['klassifikation']
        for k in klassifikationer:
            if k['@kaldenavn'] == 'Stillingsbetegnelser':
                self.create_typer(k['@uuid'],
                                  {'engagement_job_function': ['Alfabetisk']})
            if k['@kaldenavn'] == 'Rolle':
                self.create_typer(k['@uuid'],
                                  {'role_type': ['Rolle']})

            if k['@kaldenavn'] == 'Enhedstyper':
                self.create_typer(k['@uuid'],
                                  {'org_unit_type': ['Alfabetisk']})

            if k['@kaldenavn'] == 'Tilknytningstyper':
                self.create_typer(k['@uuid'],
                                  {'engagement_type': ['Alfabetisk']})

            if k['@kaldenavn'] == 'AM/MED':
                self.create_typer(k['@uuid'],
                                  {'association_type': ['Repræsentanttyper',
                                                        'Medlemstyper',
                                                        'Forbund']})

            if k['@kaldenavn'] == 'Leder':
                self.create_typer(k['@uuid'], {'responsibility': ['Ansvar'],
                                               'manager_type': ['Typer']})

            if k['@kaldenavn'] == 'SD løn enhedstyper':
                self.create_typer(k['@uuid'], {'org_unit_type':
                                               ['sd_loen_enhedstyper']})

            if k['@kaldenavn'] == 'SD løn enhedstyper':
                self.create_typer(k['@uuid'], {'time_planning':
                                               ['Tidsregistrering']})

            """
            if k['@kaldenavn'] == 'Kontaktkanaler':
                self.create_typer(k['@uuid'], {'employee_address_type':
                                               ['Lokation typer',
                                                'Egenskaber',
                                                'Engagement typer']})
            if k['@kaldenavn'] == 'Kontaktkanaler':
                self.create_typer(k['@uuid'], {'org_unit_address_type':
                                               ['Lokation typer']})
            if k['@kaldenavn'] == 'Kontaktkanaler':
                self.create_typer(k['@uuid'], {'employee_address_type':
                                               ['Engagement typer']})
            """

        if CREATE_UDVALGS_CLASSES:
            specific_klasser = [
                {'id': 'AMG', 'titel': 'AMG',
                 'facet': 'org_unit_type', 'scope': 'TEXT'},
                {'id': 'H-MED', 'titel': 'H-MED',
                 'facet': 'org_unit_type', 'scope': 'TEXT'},
                {'id': 'C-MED', 'titel': 'C-MED',
                 'facet': 'org_unit_type', 'scope': 'TEXT'},
                {'id': 'L-MED', 'titel': 'L-MED',
                 'facet': 'org_unit_type', 'scope': 'TEXT'}
            ]
        else:
            specific_klasser = []
        standard_klasser = [
            {'id': 'Lederniveau',
             'titel': 'Lederniveau',
             'facet': 'manager_level',
             'scope': 'TEXT'},
            {'id': 'EmailEmployee',
             'titel': 'Email',
             'facet': 'employee_address_type',
             'scope': 'EMAIL'},
            {'id': 'PhoneEmployee',
             'titel': 'Telefon',
             'facet': 'employee_address_type',
             'scope': 'PHONE'},
            {'id': 'AltPhoneEmployee',
             'titel': 'Alt Tlf',
             'facet': 'employee_address_type',
             'scope': 'PHONE'},
            {'id': 'p-nummer',
             'titel': 'p-nummer',
             'facet': 'org_unit_address_type',
             'scope': 'PNUMBER'},
            {'id': 'EmailUnit',
             'titel': 'Email',
             'facet': 'org_unit_address_type',
             'scope': 'EMAIL'},
            {'id': 'AddressMailUnit',
             'titel': 'Adresse',
             'facet': 'org_unit_address_type',
             'scope': 'DAR'},
            {'id': 'PhoneUnit',
             'titel': 'Telefon',
             'facet': 'org_unit_address_type',
             'scope': 'PHONE'},
            {'id': 'EAN',
             'titel': 'EAN',
             'facet': 'org_unit_address_type',
             'scope': 'EAN'},
            {'id': 'WebUnit',
             'titel': 'Webadresse',
             'facet': 'org_unit_address_type',
             'scope': 'WWW'},
            {'id': 'Ekstern',
             'titel': 'Må vises eksternt',
             'facet': 'visibility',
             'scope': 'PUBLIC'},
            {'id': 'Intern',
             'titel': 'Må vises internt',
             'facet': 'visibility',
             'scope': 'INTERNAL'},
            {'id': 'Hemmelig',
             'titel': 'Hemmelig',
             'facet': 'visibility',
             'scope': 'SECRET'},
            {'id': 'Teammedarbejder',
             'titel': 'Teammedarbejder',
             'facet': 'association_type',
             'scope': 'TEXT'},
            {'id': 'Projektleder',
             'titel': 'Projektleder',
             'facet': 'association_type',
             'scope': 'TEXT'},
            {'id': 'Projektgruppemedlem',
             'titel': 'Projektgruppemedlem',
             'facet': 'association_type',
             'scope': 'TEXT'},
            # Last three possbily not general to all APOS imports
            {'id': 'Afdelingskode',
             'titel': 'Afdelingskode',
             'facet': 'org_unit_address_type',
             'scope': 'TEXT'},
            {'id': 'Formaalskode',
             'titel': 'Formålskode',
             'facet': 'org_unit_address_type',
             'scope': 'TEXT'},
            {'id': 'Skolekode',
             'titel': 'Skolekode',
             'facet': 'org_unit_address_type',
             'scope': 'TEXT'},
        ]

        for klasse in specific_klasser + standard_klasser:
            self.importer.add_klasse(identifier=klasse['id'],
                                     title=klasse['titel'],
                                     user_key=klasse['id'],
                                     scope=klasse['scope'],
                                     facet_type_ref=klasse['facet'])

    def _read_ous_from_apos(self, org_uuid):
        url = "app-organisation/GetEntireHierarchy?uuid={}"
        r = self._apos_lookup(url.format(org_uuid))
        org_units = r['node']
        return org_units

    def _create_ou_from_apos(self, apos_unit, parent=None, enhedstype=None):
        """ Create a MO org_unit from the corresponding APOS object,
        includes looking up address information """
        url = "app-organisation/GetUnitDetails?uuid={}"
        r = self._apos_lookup(url.format(apos_unit['@uuid']))
        details = r['enhed']

        """
        The handling of apos integration-data might turn out to be specfic for
        Ballerup in the current  implementation. Once other imports also needs these
        values, we should make a parameter list.
        """

        # Default i Ballerup:
        time_planning = '41504f53-0203-0028-4158-41504f494e54'
        if details['opgaver']:
            opgaver = details['opgaver']['opgave']
            if not isinstance(opgaver, list):
                opgaver = [opgaver]

            for opgave in opgaver:
                if opgave['@uuid'] in ('41504f53-0203-0028-4158-41504f494e54',
                                       '41504f53-0203-0029-4158-41504f494e54',
                                       '41504f53-0203-0030-4158-41504f494e54'):
                    time_planning = opgave['@uuid']

        unit_id = int(apos_unit['@objectid'])
        if 'overordnet' in details:
            gyldighed = {'@fra': details['overordnet']['@fra'],
                         '@til': details['overordnet']['@til']
                         }
            fra, til = _format_time(gyldighed)
        else:
            fra, til = _format_time(details['gyldighed'])

        enhedstype = details['@enhedstype']

        # NY7
        if enhedstype == '58380fd3-b3fb-4f84-a56f-15c9716972c1':
            enhedstype = '42b9042f-5f20-4998-b0e5-c4deb6c5f42e'

        # NY6
        if enhedstype == '25eebd5c-d774-469c-97af-f8d9ca2780c9':
            enhedstype = '414a035e-9c22-42eb-b035-daa7d7f2ade8'

        # NY5
        if enhedstype == 'b4bd5908-c724-4544-bff3-324832156ff3':
            enhedstype = '819ae28e-04e0-4030-880e-7b699faeaff9'

        # NY4
        if enhedstype == 'f16e9d0c-2692-43d3-8def-62c0c7a75fdf':
            enhedstype = 'ff8c3f53-85ec-44d7-a9d6-07c619ac50df'

        # NY3
        if enhedstype == '2c73e9b0-785d-48a4-b14b-656c596be759':
            enhedstype = '70c69826-4ba1-4e1e-82f0-4c47c89a7ecc'

        # NY2
        if enhedstype == '68fea789-87ce-451c-8ad7-8b7525c26978':
            enhedstype = 'ec882c49-3cc2-4bc9-994f-a6f29136401b'

        # NY1
        if enhedstype == '5cbde7bf-7632-4b6a-ae51-691a89d8df7a':
            enhedstype = 'd9bd186b-3c11-4dbf-92d1-4e3b61140302'

        # Afdelingsniveau
        if enhedstype in ('81fc2195-d782-4ae0-8ab7-56a1f7ad92e2',
                          '58380fd3-b3fb-4f84-a56f-15c9716972c1'):
            enhedstype = '345a8893-eb1f-4e20-b76d-63b95b5809f6'

        apos_int = {}
        if 'integrationAttributter' in details:
            if details['integrationAttributter'] is not None:
                attributter = details['integrationAttributter']['attribut']
                if not isinstance(attributter, list):
                    attributter = [attributter]

                for attribut in attributter:
                    if isinstance(attribut, str):
                        continue

                    if len(attribut['@vaerdi']) > 0:
                        apos_int[attribut['@navn']] = attribut['@vaerdi']

        if 'Afdelingskode' in apos_int:
            user_key = apos_int['Afdelingskode']
            apos_int['Afdelingskode'] = apos_unit['@brugervendtNoegle']
        else:
            user_key = apos_unit['@brugervendtNoegle']

        unit = self.importer.add_organisation_unit(
            identifier=unit_id,
            uuid=apos_unit['@uuid'],
            name=apos_unit['@navn'],
            user_key=user_key,
            time_planning_ref=time_planning,
            type_ref=enhedstype,
            date_from=fra,
            date_to=til,
            parent_ref=parent)

        for key, value in apos_int.items():
            logger.debug('Add address type {}: value {}'.format(key, value))
            self.importer.add_address_type(
                organisation_unit=unit_id,
                type_ref=key,
                value=value,
                date_from=fra
            )

        if apos_unit['@uuid'] in self.ean:
            self.importer.add_address_type(
                organisation_unit=unit_id,
                type_ref='EAN',
                value=self.ean[apos_unit['@uuid']],
                date_from=fra
            )

        # This is most likely the orgunit relation information
        # print(details['tilknyttedeEnheder']))

        locations = self.read_locations(apos_unit)
        for location in locations:
            if location['pnummer']:
                try:
                    self.importer.add_address_type(
                        organisation_unit=unit_id,
                        type_ref='p-nummer',
                        value=location['pnummer'],
                        date_from=fra)
                except AssertionError:  # pnumber added multiple times
                    pass

            if location['dawa_uuid']:
                try:
                    self.importer.add_address_type(
                        organisation_unit=unit_id,
                        type_ref='AddressMailUnit',
                        value=location['dawa_uuid'],
                        date_from=fra)
                except AssertionError:  # Address already added
                    pass
        return unit

    def create_associations_for_ou(self, unit):
        """ Returnerer tilsyneladende personer som ikke er ansat.
        Dette skal undersges.
        """
        url = "app-organisation/GetAttachedPersonsForUnit?uuid={}"
        uuid = self.object_to_uuid[unit]
        added_persons = []
        associations = self._apos_lookup(url.format(uuid))
        if int(associations['total']) > 0:
            persons = associations['person']
            if not isinstance(persons, list):
                persons = [persons]

            for person in persons:
                for p in person.values():
                    if p in added_persons:
                        # Some persons are reported more than once?
                        break
                    else:
                        added_persons.append(p)
                    if not self.importer.check_if_exists('employee', p):
                        logger.warning('Association error: {}'.format(p))
                        break

                    details = self.importer.get_details(owner_type="employee",
                                                        owner_ref=p,
                                                        type_id="engagement")
                    from_date = details[0].date_from

                    self.importer.add_association(
                        employee=p,
                        organisation_unit=unit,
                        association_type_ref=ANSAT_UUID,
                        date_from=from_date
                    )

    def update_contact_information(self, employee):
        kontakt = employee['klassifikationKontaktKanaler']
        if kontakt is None:
            return
        kontaktmuligheder = kontakt['klassifikationKontaktKanal']
        for kontaktmulighed in kontaktmuligheder:
            if isinstance(kontaktmulighed, str):
                return
            value = kontaktmulighed['@vaerdi']
            if value:
                apos_type = kontaktmulighed['@type']

                # klasse = self.importer.get('klasse', apos_type)
                employee_identifier = employee['person']['@uuid']
                if apos_type == EMAIL_NAME:
                    klasse_ref = 'EmailEmployee'
                elif apos_type == MAIN_PHONE_NAME:
                    klasse_ref = 'PhoneEmployee'
                elif apos_type == ALT_PHONE_NAME:
                    klasse_ref = 'AltPhoneEmployee'
                else:  # This should never happen
                    logger.error('Ukendt kontaktmulighed: {}'.format(apos_type))
                    raise Exception('Ukendt kontaktmulighed')
                try:
                    fra, til = _format_time(employee['gyldighed'])

                    self.importer.add_address_type(
                        employee=employee_identifier,
                        value=value,
                        type_ref=klasse_ref,
                        # date_from=GLOBAL_DATE,
                        # date_to=None
                        date_from=fra,
                        date_to=til
                    )
                except AssertionError:
                    pass  # Already inserted

    def update_tasks(self, employee, unit):
        tasks = employee['opgaver']['opgave']
        fra, til = _format_time(employee['gyldighed'])
        if isinstance(tasks, list):
            assert(len(tasks) == 2)
            assert(tasks[0]['@klassifikation'] == 'stillingsbetegnelser')
            assert(tasks[1]['@klassifikation'] == 'tilknytningstyper')
            stilling = tasks[0]['@uuid']
            tilknytning = tasks[1]['@uuid']
            if not tilknytning == '56e1214a-330f-4592-89f3-ae3ee8d5b2e6':
                # Tilknyting bør altid være 'Ansat', vi ignorerer # derfor
                # denne oplysning, men kan debugge her hvis der opstår behov
                pass
        else:
            assert(tasks['@klassifikation'] == 'stillingsbetegnelser')
            stilling = tasks['@uuid']

        if not (
                self.importer.check_if_exists('klasse', stilling) and
                self.importer.check_if_exists('organisation_unit', unit) and
                self.importer.check_if_exists('employee',
                                              employee['person']['@uuid'])
        ):
            logger.error('Medarbejder ukendt: {}'.format(employee))
            1/0

        engagement_ref = '56e1214a-330f-4592-89f3-ae3ee8d5b2e6'  # Ansat
        logger.debug(
            'Add engagement, person: {}, from: {}, to: {}'.format(
                employee['person']['@uuid'], fra, til
            )
        )

        self.importer.add_engagement(
            employee=employee['person']['@uuid'],
            uuid=employee['@uuid'],
            user_key=employee['@uuid'],
            organisation_unit=unit,
            job_function_ref=stilling,
            engagement_type_ref=engagement_ref,
            date_from=fra,
            date_to=til)

    def create_employees_for_ou(self, unit):
        url = 'composite-services/GetEngagementDetailed?unitUuid={}'
        medarbejdere = self._apos_lookup(url.format(unit))

        if 'total' not in medarbejdere:
            return   # This unit has no employees
        if medarbejdere['total'] == '0':
            return
        elif medarbejdere['total'] == '1':
            medarbejdere = [medarbejdere['engagementer']['engagement']]
        else:
            medarbejdere = medarbejdere['engagementer']['engagement']

        # Be carefull, the employee has uuids both as persons and as
        # employees
        for medarbejder in medarbejdere:
            objectid = int(medarbejder['pathToRoot']['enhed'][-1]['@objectid'])
            person = medarbejder['person']
            given_name = person['@fornavn']
            if person['@mellemnavn']:
                given_name = given_name + ' ' + person['@mellemnavn']
            sur_name = person['@efternavn']
            fra, til = _format_time(medarbejder['gyldighed'])
            bvn = medarbejder['@brugervendtNoegle']

            if not self.importer.check_if_exists('employee', person['@uuid']):
                logger.info('New employee: {}'.format(person))

                name = (given_name, sur_name)
                logger.info('Add employee: {}, user_key: {}'.format(name, bvn))
                self.importer.add_employee(name=name,
                                           uuid=person['@uuid'],
                                           identifier=person['@uuid'],
                                           cpr_no=person['@personnummer'],
                                           user_key=bvn)

            self.update_contact_information(medarbejder)
            self.update_tasks(medarbejder, objectid)

    def get_ou_functions(self, unit):
        url = 'app-organisation/GetFunctionsForUnit?uuid={}'
        uuid = self.object_to_uuid[unit]
        org_funcs = self._apos_lookup(url.format(uuid))

        if int(org_funcs['total']) == 0:
            return
        for func in org_funcs['function']:
            if not isinstance(func, collections.OrderedDict):
                continue  # This func is empty

            fra, til = _format_time(func['gyldighed'])
            apos_persons = func['persons']
            if not apos_persons:
                continue
            if not apos_persons['person']:
                continue  # Vacant manager?

            personer = apos_persons['person']
            if not isinstance(personer, list):
                personer = [personer]

            # Only one person should ever be in this list, but sometimes
            # empty entries are also included. These are filtered out by
            # abcense of a person uuid.
            for person in personer:
                if not person['@uuid']:
                    continue

                assert(func['units']['unit']['@uuid'] == uuid)
                tasks = func['tasks']['task']

                manager_type = None
                manager_responsibility = []

                for task in tasks:
                    try:
                        klasse = task['@uuid']
                    except TypeError:
                        continue

                    """ This is the old procedure, check the new one works
                    try:  # We have a few problematic Klasser, chack manually
                        self.importer.get('klasse', klasse)
                    except KeyError:
                        self.klassifikation_errors[klasse] = True
                        continue
                    """
                    klasse_ref = self.importer.get('klasse', klasse)
                    # We have a few problematic Klasser, chack manually
                    if klasse_ref is None:
                        print(klasse)
                        self.klassifikation_errors[klasse] = True
                        continue

                    facet = klasse_ref.facet_type_ref

                    if facet == 'manager_type':

                        manager_type = klasse
                    elif facet == 'responsibility':
                        manager_responsibility.append(klasse)
                    elif facet in ('engagement_job_function', 'engagement_type'):
                        pass
                    else:
                        logger.error('Unkown facet: {}'.format(facet))

                if manager_type:
                    try:
                        logger.debug('Add manager: {}'.format(person['@uuid']))
                        self.importer.add_manager(
                            employee=person['@uuid'],
                            organisation_unit=unit,
                            manager_level_ref='Lederniveau',
                            address_uuid=None,  # TODO?
                            manager_type_ref=manager_type,
                            responsibility_list=manager_responsibility,
                            uuid=func['@uuid'],
                            date_from=fra,
                            date_to=til
                        )
                    except ReferenceError:
                        logger.warning(
                            'Problem adding manager: {}'.format(person['@uuid'])
                        )

    def create_ou_tree(self, org_uuid, enhedstype=None):
        org_units = self._read_ous_from_apos(org_uuid)
        nodes = {}
        # The root org is always first row in APOS
        objectid = int(org_units[0]['@objectid'])
        self.object_to_uuid[objectid] = org_units[0]['@uuid']
        nodes[objectid] = self._create_ou_from_apos(org_units[0])
        del(org_units[0])

        while org_units:
            remaining_org_units = []
            new = {}
            for unit in org_units:
                over_id = int(unit['@overordnetid'])
                unit_id = int(unit['@objectid'])
                if over_id in nodes.keys():
                    self.object_to_uuid[unit_id] = unit['@uuid']
                    new[unit_id] = self._create_ou_from_apos(
                        unit,
                        over_id,
                        enhedstype=enhedstype)
                    self.create_employees_for_ou(unit['@uuid'])
                else:
                    remaining_org_units.append(unit)
            logger.debug('Remaining org units: {}'.format(len(remaining_org_units)))
            org_units = remaining_org_units
            nodes.update(new)

    def create_managers_and_associatins(self):
        """ Create org_funcs, at the momen this means managers """
        # units = self.org.OrganisationUnit.export()
        units = self.importer.export('organisation_unit')
        for unit in units:
            self.get_ou_functions(unit)
            self.create_associations_for_ou(unit)
        logger.info('Finished creating managers')

    def add_all_missing_employees(self):
        """
        Call this function to retrive any person known by apos and add them to MO
        if they are not already imported at this point.
        """
        url = 'app-part/GetPersonList'
        persons = self._apos_lookup(url)

        for person in persons['person']:
            given_name = person['@fornavn'] + ' '
            if person['@mellemnavn']:
                given_name += person['@mellemnavn']
            sur_name = person['@efternavn']
            name = (given_name, sur_name)

            cpr = person['@personnummer']
            if not len(cpr) == 10:
                # print('Unable to import {}'.format(person))
                continue

            if not self.importer.check_if_exists('employee', person['@uuid']):
                logger.info('New employee: {}'.format(person))
                self.importer.add_employee(
                    name=name,
                    uuid=person['@uuid'],
                    identifier=person['@uuid'],
                    cpr_no=person['@personnummer'],
                    user_key=person['@uuid']
                )
        logger.info('Finished importig remaining employees')



import unittest
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Optional
from unittest.mock import patch

from helpers import dummy_settings

from integrations.os2sync.os2mo import get_sts_user
from integrations.os2sync.os2mo import try_get_ad_user_key

uuid = "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b"
mo_employee_url_end = "e/{}/".format(uuid)
mo_employee_address_url_end = mo_employee_url_end + "details/address"
mo_employee_engagement_url_end = (
    mo_employee_url_end + "details/engagement?calculate_primary=true"
)
mo_employee_it_url_end = mo_employee_url_end + "details/it"

mo_employee = {
    "cpr_no": "0602602389",
    "givenname": "Solveig",
    "name": "Solveig Kuhlenhenke",
    "nickname": "",
    "nickname_givenname": "",
    "nickname_surname": "",
    "org": {
        "name": "Kolding Kommune",
        "user_key": "Kolding Kommune",
        "uuid": "3b866d97-0b1f-48e0-8078-686d96f430b3",
    },
    "surname": "Kuhlenhenke",
    "user_key": "SolveigK_user_key",
    "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
}

mo_employee_address = [
    {
        "address_type": {
            "example": None,
            "facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "full_name": "Lokation",
            "name": "Lokation",
            "owner": None,
            "scope": "TEXT",
            "top_level_facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "user_key": "LocationEmployee",
            "uuid": "d34e2ee9-ec64-4259-a315-f681c3417866",
        },
        "href": None,
        "name": "Bygning 15",
        "person": {
            "givenname": "Solveig",
            "name": "Solveig Kuhlenhenke",
            "nickname": "",
            "nickname_givenname": "",
            "nickname_surname": "",
            "surname": "Kuhlenhenke",
            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        },
        "user_key": "Bygning 15",
        "uuid": "1940ecde-c1ad-418a-bf15-194df92a0b2b",
        "validity": {"from": "2003-08-13", "to": None},
        "value": "Bygning 15",
    },
    {
        "address_type": {
            "example": None,
            "facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "full_name": "Postadresse",
            "name": "Postadresse",
            "owner": None,
            "scope": "DAR",
            "top_level_facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "user_key": "AdressePostEmployee",
            "uuid": "e75f74f5-cbc4-4661-b9f4-e6a9e05abb2d",
        },
        "href": "https://www.openstreetmap.org/?mlon=9.40634869&mlat=55.55868524&zoom=16",
        "name": "Bakkedraget 28, Vester Nebel, 6040 Egtved",
        "person": {
            "givenname": "Solveig",
            "name": "Solveig Kuhlenhenke",
            "nickname": "",
            "nickname_givenname": "",
            "nickname_surname": "",
            "surname": "Kuhlenhenke",
            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        },
        "user_key": "0a3f50ba-66b3-32b8-e044-0003ba298018",
        "uuid": "544b4172-63db-4b47-8322-fa0ab4a42de0",
        "validity": {"from": "2003-08-13", "to": None},
        "value": "0a3f50ba-66b3-32b8-e044-0003ba298018",
    },
    {
        "address_type": {
            "example": None,
            "facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "full_name": "Telefon",
            "name": "Telefon",
            "owner": None,
            "scope": "PHONE",
            "top_level_facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "user_key": "PhoneEmployee",
            "uuid": "05b69443-0c9f-4d57-bb4b-a8c719afff89",
        },
        "href": "tel:55677853",
        "name": "55677853",
        "person": {
            "givenname": "Solveig",
            "name": "Solveig Kuhlenhenke",
            "nickname": "",
            "nickname_givenname": "",
            "nickname_surname": "",
            "surname": "Kuhlenhenke",
            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        },
        "user_key": "55677853",
        "uuid": "6293d387-7504-4b04-b32e-49d93140acf1",
        "validity": {"from": "2003-08-13", "to": None},
        "value": "55677853",
        "visibility": {
            "example": None,
            "name": "Hemmelig",
            "owner": None,
            "scope": "SECRET",
            "user_key": "Hemmelig",
            "uuid": "baaabba6-829c-44bf-ad45-db9d9c72f834",
        },
    },
    {
        "address_type": {
            "example": None,
            "facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "full_name": "Email",
            "name": "Email",
            "owner": None,
            "scope": "EMAIL",
            "top_level_facet": {
                "description": "",
                "user_key": "employee_address_type",
                "uuid": "5b3a55b1-958c-416e-9054-606b2c9e4fcd",
            },
            "user_key": "EmailEmployee",
            "uuid": "f376deb8-4743-4ca6-a047-3241de8fe9d2",
        },
        "href": "mailto:solveigk@kolding.dk",
        "name": "solveigk@kolding.dk",
        "person": {
            "givenname": "Solveig",
            "name": "Solveig Kuhlenhenke",
            "nickname": "",
            "nickname_givenname": "",
            "nickname_surname": "",
            "surname": "Kuhlenhenke",
            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        },
        "user_key": "solveigk@kolding.dk",
        "uuid": "6c454f09-a2d0-4904-b64d-3ee90dfbfc8a",
        "validity": {"from": "2003-08-13", "to": None},
        "value": "solveigk@kolding.dk",
    },
]

mo_employee_engagement = [
    {
        "engagement_type": {
            "example": None,
            "facet": {
                "description": "",
                "user_key": "engagement_type",
                "uuid": "182df2a8-2594-4a3f-9103-a9894d5e0c36",
            },
            "full_name": "Ansat",
            "name": "Ansat",
            "owner": None,
            "scope": "TEXT",
            "top_level_facet": {
                "description": "",
                "user_key": "engagement_type",
                "uuid": "182df2a8-2594-4a3f-9103-a9894d5e0c36",
            },
            "user_key": "Ansat",
            "uuid": "8acc5743-044b-4c82-9bb9-4e572d82b524",
        },
        "extension_1": None,
        "extension_10": None,
        "extension_2": None,
        "extension_3": None,
        "extension_4": None,
        "extension_5": None,
        "extension_6": None,
        "extension_7": None,
        "extension_8": None,
        "extension_9": None,
        "fraction": None,
        "integration_data": {
            "Artificial import": '"be8300fe7fda60be2c1e57634dd56'
            '7d8cb8977aa5a065cace7d0a1ae03910981"STOP_DUMMY'
        },
        "is_primary": None,
        "job_function": {
            "example": None,
            "facet": {
                "description": "",
                "user_key": "engagement_job_function",
                "uuid": "82155433-489f-4ddc-84f6-89911a454c7c",
            },
            "full_name": "Bogopsætter",
            "name": "Bogopsætter",
            "owner": None,
            "scope": "TEXT",
            "top_level_facet": {
                "description": "",
                "user_key": "engagement_job_function",
                "uuid": "82155433-489f-4ddc-84f6-89911a454c7c",
            },
            "user_key": "Bogopsætter",
            "uuid": "cf84f415-a6bd-4b4d-9b06-91ea392a8543",
        },
        "org_unit": {
            "name": "Kolding Kommune",
            "user_key": "Kolding Kommune",
            "uuid": "f06ee470-9f17-566f-acbe-e938112d46d9",
            "validity": {"from": "1960-01-01", "to": None},
        },
        "person": {
            "givenname": "Solveig",
            "name": "Solveig Kuhlenhenke",
            "nickname": "",
            "nickname_givenname": "",
            "nickname_surname": "",
            "surname": "Kuhlenhenke",
            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        },
        "primary": {
            "example": None,
            "facet": {
                "description": "",
                "user_key": "primary_type",
                "uuid": "4afba2d2-4846-42df-981b-0a85ee8de8db",
            },
            "full_name": "Primær",
            "name": "Primær",
            "owner": None,
            "scope": "3000",
            "top_level_facet": {
                "description": "",
                "user_key": "primary_type",
                "uuid": "4afba2d2-4846-42df-981b-0a85ee8de8db",
            },
            "user_key": "primary",
            "uuid": "0644cd06-b84b-42e0-95fe-ce131c21fbe6",
        },
        "user_key": "-",
        "uuid": "94f04266-f744-4c06-8f4e-9561abfbff75",
        "validity": {"from": "2003-08-13", "to": None},
    }
]

mo_employee_it = [
    {
        "itsystem": {
            "name": "OpenDesk",
            "reference": None,
            "system_type": None,
            "user_key": "OpenDesk",
            "uuid": "db519bfd-0fdd-4e5d-9337-518d1dbdbfc9",
            "validity": {"from": "1900-01-01", "to": None},
        },
        "org_unit": None,
        "person": {
            "givenname": "Solveig",
            "name": "Solveig Kuhlenhenke",
            "nickname": "",
            "nickname_givenname": "",
            "nickname_surname": "",
            "surname": "Kuhlenhenke",
            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        },
        "user_key": "SolveigKOpenDesk",
        "uuid": "a029d0ac-7ea6-4023-8680-57825e43af8c",
        "validity": {"from": "2003-08-13", "to": None},
    },
    {
        "itsystem": {
            "name": "Active Directory",
            "reference": None,
            "system_type": None,
            "user_key": "Active Directory",
            "uuid": "a1608e69-c422-404f-a6cc-b873c50af111",
            "validity": {"from": "1900-01-01", "to": None},
        },
        "org_unit": None,
        "person": {
            "givenname": "Solveig",
            "name": "Solveig Kuhlenhenke",
            "nickname": "",
            "nickname_givenname": "",
            "nickname_surname": "",
            "surname": "Kuhlenhenke",
            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        },
        "user_key": "SolveigK_AD_logon",
        "uuid": "a2fb2581-c57a-46ad-8a21-30118a3859b7",
        "validity": {"from": "2003-08-13", "to": None},
    },
]


class MockResponse:
    def __init__(self, value):
        self.value = value

    def raise_for_status(self):
        pass

    def json(self):
        return deepcopy(self.value)


def patched_session_get(url: str, params: Optional[Dict[Any, Any]] = None, **kwargs):
    if params:
        raise ValueError("unexpected params: {}".format(params))
    if kwargs:
        raise ValueError("unexpected kwargs: {}".format(kwargs))
    for legal_end, value in {
        mo_employee_url_end: mo_employee,
        mo_employee_it_url_end: mo_employee_it,
        mo_employee_address_url_end: mo_employee_address,
        mo_employee_engagement_url_end: mo_employee_engagement,
    }.items():
        if url.endswith(legal_end):
            return MockResponse(value)

    raise ValueError("unexpected url: {}".format(url))


class TestsMOAd(unittest.TestCase):
    @patch("integrations.os2sync.os2mo.os2mo_get", patched_session_get)
    def test_get_ad_user_key(self):
        expected = "SolveigK_AD_logon"
        self.assertEqual(expected, try_get_ad_user_key(uuid))

    @patch("integrations.os2sync.os2mo.os2mo_get", patched_session_get)
    def test_mo_client_default(self):
        expected = {
            "Email": "solveigk@kolding.dk",
            "Person": {"Cpr": "0602602389", "Name": "Solveig Kuhlenhenke"},
            "Positions": [],
            "UserId": "SolveigK_AD_logon",
            "Uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
        }
        settings = dummy_settings
        settings.os2sync_xfer_cpr = True
        self.assertEqual(expected, get_sts_user(uuid, [], settings=settings))



import json
import pathlib

from integrations import cpr_mapper
from os2mo_data_import import ImportHelper
from integrations.SD_Lon import sd_importer
from integrations.ad_integration import ad_reader

cfg_file = pathlib.Path.cwd() / 'settings' / 'settings.json'
if not cfg_file.is_file():
    raise Exception('No setting file')
settings = json.loads(cfg_file.read_text())

cpr_map = pathlib.Path.cwd() / 'settings' / 'cpr_uuid_map.csv'
if not cpr_map.is_file():
    raise Exception('No mapping file')
employee_mapping = cpr_mapper.employee_mapper(str(cpr_map))

importer = ImportHelper(
    create_defaults=True,
    mox_base=settings['mox.base'],
    mora_base=settings['mora.base'],
    store_integration_data=False,
    seperate_names=True
)

ad_reader = ad_reader.ADParameterReader()

sd = sd_importer.SdImport(
    importer,
    ad_info=ad_reader,
    employee_mapping=employee_mapping
)

sd.create_ou_tree(create_orphan_container=True)
sd.create_employees()

importer.import_all()



import uuid
from collections import OrderedDict
from datetime import date
from datetime import timedelta
from unittest.mock import call
from unittest.mock import MagicMock
from unittest.mock import patch

import click
import hypothesis.strategies as st
import pytest
from hypothesis import given
from parameterized import parameterized
from ra_utils.attrdict import attrdict
from ra_utils.generate_uuid import uuid_generator

from .fixtures import get_employment_fixture
from .fixtures import get_read_employment_changed_fixture
from .fixtures import get_sd_person_fixture
from .fixtures import read_employment_fixture
from integrations.SD_Lon.config import ImporterSettings
from integrations.SD_Lon.sd_changed_at import ChangeAtSD
from integrations.SD_Lon.sd_changed_at import get_from_date
from test_case import DipexTestCase


@given(test_from_date=st.datetimes())
def test_getfrom_date(test_from_date):
    """Test reading latest date from rundb"""
    with patch(
        "integrations.rundb.db_overview.DBOverview._read_last_line",
        return_value=((test_from_date, "Update ended at ---")),
    ):
        from_date = get_from_date("test", force=False)
        assert from_date == test_from_date


@given(test_from_date=st.datetimes())
def test_getfrom_date_running(test_from_date):
    """Test raising error if last import didn't finish"""
    with patch(
        "integrations.rundb.db_overview.DBOverview._read_last_line",
        return_value=((test_from_date, "Running")),
    ):
        with pytest.raises(click.ClickException):
            get_from_date("test", force=False)


@given(test_from_date=st.datetimes())
@patch("integrations.rundb.db_overview.DBOverview.delete_last_row")
def test_getfrom_date_force(delete_mock, test_from_date):
    """Test reading though last import didn't finish using force=True"""
    with patch(
        "integrations.rundb.db_overview.DBOverview._read_last_line",
        return_value=((test_from_date, "Running")),
    ):
        get_from_date("test", force=True)
        delete_mock.assert_called_once()


class ChangeAtSDTest(ChangeAtSD):
    def __init__(self, *args, **kwargs):
        self.morahelper_mock = MagicMock()
        self.morahelper_mock.read_organisation.return_value = "org_uuid"
        self.primary_types_mock = MagicMock()
        self.primary_engagement_mock = MagicMock()
        self.fix_departments_mock = MagicMock()

        self._get_job_sync = MagicMock()

        self._create_class = MagicMock()
        self._create_class.return_value = "new_class_uuid"

        super().__init__(*args, **kwargs)

    def _get_primary_types(self, mora_helper):
        return self.primary_types_mock

    def _get_primary_engagement_updater(self):
        return self.primary_engagement_mock

    def _get_fix_departments(self):
        return self.fix_departments_mock

    def _read_forced_uuids(self):
        return {}

    def _get_mora_helper(self, mora_base):
        return self.morahelper_mock


def setup_sd_changed_at(updates=None, hours=24):
    # TODO: remove integrations.SD_Lon.terminate_engagement_with_to_only
    settings_dict = {
        "municipality_code": 100,
        "municipality_name": "Pladderballe Kommune",
        "sd_global_from_date": "1970-01-01",
        "sd_import_run_db": "run_db.sqlite",
        "sd_institution_identifier": "XY",
        "sd_password": "secret",
        "sd_user": "user",
        "sd_job_function": "JobPositionIdentifier",
        "sd_use_ad_integration": False,
        "sd_monthly_hourly_divide": 8000,
        "mora_base": "http://dummy.url",
        "mox_base": "http://dummy.url",
        "sd_terminate_engagement_with_to_only": False,
    }
    if updates:
        settings_dict.update(updates)

    settings = ImporterSettings.parse_obj(settings_dict)

    today = date.today()
    start_date = today

    sd_updater = ChangeAtSDTest(
        settings, start_date, start_date + timedelta(hours=hours)
    )

    return sd_updater


class Test_sd_changed_at(DipexTestCase):
    @patch("integrations.SD_Lon.sd_common.sd_lookup_settings")
    @patch("integrations.SD_Lon.sd_common._sd_request")
    def test_get_sd_person(self, sd_request, sd_settings):
        """Test that read_person does the expected transformation."""
        sd_settings.return_value = ("", "", "")

        cpr = "0101709999"
        sd_reply, expected_read_person_result = get_sd_person_fixture(
            cpr=cpr, first_name="John", last_name="Deere", employment_id="01337"
        )

        sd_request.return_value = sd_reply

        sd_updater = setup_sd_changed_at()
        result = sd_updater.get_sd_person(cpr=cpr)
        self.assertEqual(result, expected_read_person_result)

    def test_update_changed_persons(self):

        cpr = "0101709999"
        first_name = "John"
        last_name = "Deere"

        _, read_person_result = get_sd_person_fixture(
            cpr=cpr,
            first_name=first_name,
            last_name=last_name,
            employment_id="01337",
        )

        sd_updater = setup_sd_changed_at()
        sd_updater.get_sd_person = lambda cpr: read_person_result

        generate_uuid = uuid_generator("test")
        org_uuid = str(generate_uuid("org_uuid"))
        user_uuid = str(generate_uuid("user_uuid"))

        sd_updater.org_uuid = org_uuid

        morahelper = sd_updater.morahelper_mock
        morahelper.read_user.return_value = {
            "uuid": user_uuid,
            "name": " ".join(["Old firstname", last_name]),
            "first_name": "Old firstname",
            "surname": last_name,
        }

        _mo_post = morahelper._mo_post
        _mo_post.return_value = attrdict(
            {"status_code": 201, "json": lambda: user_uuid}
        )
        self.assertFalse(_mo_post.called)
        sd_updater.update_changed_persons(in_cpr=cpr)
        _mo_post.assert_called_with(
            "e/create",
            {
                "type": "employee",
                "givenname": first_name,
                "surname": last_name,
                "cpr_no": cpr,
                "org": {"uuid": org_uuid},
                "uuid": user_uuid,
                "user_key": user_uuid,
            },
        )

    @given(status=st.sampled_from(["1", "S"]))
    @patch("integrations.SD_Lon.sd_common.sd_lookup_settings")
    @patch("integrations.SD_Lon.sd_common._sd_request")
    def test_read_employment_changed(
        self,
        sd_request,
        sd_settings,
        status,
    ):
        sd_settings.return_value = ("", "", "")

        sd_reply, expected_read_employment_result = read_employment_fixture(
            cpr="0101709999",
            employment_id="01337",
            job_id="1234",
            job_title="EDB-Mand",
            status=status,
        )

        sd_request.return_value = sd_reply
        sd_updater = setup_sd_changed_at()
        result = sd_updater.read_employment_changed()
        self.assertEqual(result, expected_read_employment_result)

    def test_do_not_create_engagement_for_inconsistent_external_emp(self):
        """
        We are testing bullet 4 in
        https://os2web.atlassian.net/browse/MO-245, i.e. that we do not
        create a MO engagement for a newly created external SD employee
        who (unintentionally) has a JobPositionIdentifier below
        no_salary_minimum.

        NOTE: an external SD employee has an EmploymentIdentifier containing
        letters (at least in some municipalities)
        """

        sd_updater = setup_sd_changed_at(
            {
                "sd_no_salary_minimum_id": 9000,
            }
        )
        sd_updater.read_employment_changed = (
            lambda: get_read_employment_changed_fixture(
                employment_id="ABCDE", job_pos_id=8000  # See doc-string above
            )
        )

        morahelper = sd_updater.morahelper_mock
        morahelper.read_user.return_value.__getitem__.return_value = "user_uuid"

        sd_updater.create_new_engagement = MagicMock()

        # Act
        sd_updater.update_all_employments()

        # Assert
        sd_updater.create_new_engagement.assert_not_called()

    @given(status=st.sampled_from(["1", "S"]))
    def test_update_all_employments(self, status):

        cpr = "0101709999"
        employment_id = "01337"

        _, read_employment_result = read_employment_fixture(
            cpr=cpr,
            employment_id=employment_id,
            job_id="1234",
            job_title="EDB-Mand",
            status=status,
        )

        sd_updater = setup_sd_changed_at()
        sd_updater.read_employment_changed = lambda: read_employment_result

        morahelper = sd_updater.morahelper_mock
        morahelper.read_user.return_value.__getitem__.return_value = "user_uuid"

        if status == "1":  # Creates call create engagement, if no eng exists
            sd_updater.create_new_engagement = MagicMock()

            engagement = read_employment_result[0]["Employment"]
            # First employment status entry from read_employment_result
            status = read_employment_result[0]["Employment"]["EmploymentStatus"][0]

            self.assertFalse(sd_updater.create_new_engagement.called)
            sd_updater.update_all_employments()
            sd_updater.create_new_engagement.assert_called_with(
                engagement, status, cpr, "user_uuid"
            )
        elif status == "S":  # Deletes call terminante engagement
            morahelper.read_user_engagement.return_value = [{"user_key": employment_id}]
            sd_updater._terminate_engagement = MagicMock()

            status = read_employment_result[0]["Employment"]["EmploymentStatus"]

            self.assertFalse(sd_updater._terminate_engagement.called)
            sd_updater.update_all_employments()
            sd_updater._terminate_engagement.assert_called_with(
                user_key=employment_id,
                person_uuid="user_uuid",
                from_date=status["ActivationDate"],
            )

    @parameterized.expand(
        [
            ["07777", "monthly pay"],
            ["90000", "hourly pay"],
            ["C3-P0", "employment pay"],
        ]
    )
    def test_create_new_engagement(self, employment_id, engagement_type):

        cpr = "0101709999"
        job_id = "1234"

        _, read_employment_result = read_employment_fixture(
            cpr=cpr,
            employment_id=employment_id,
            job_id=job_id,
            job_title="EDB-Mand",
        )

        sd_updater = setup_sd_changed_at()

        sd_updater.engagement_types = {
            "månedsløn": "monthly pay",
            "timeløn": "hourly pay",
            "engagement_type" + job_id: "employment pay",
        }

        morahelper = sd_updater.morahelper_mock

        # Load noop NY logic
        sd_updater.apply_NY_logic = (
            lambda org_unit, user_key, validity, person_uuid: org_unit
        )
        # Set primary types
        sd_updater.primary_types = {
            "primary": "primary_uuid",
            "non_primary": "non_primary_uuid",
            "no_salary": "no_salary_uuid",
            "fixed_primary": "fixed_primary_uuid",
        }

        engagement = read_employment_result[0]["Employment"]
        # First employment status entry from read_employment_result
        status = read_employment_result[0]["Employment"]["EmploymentStatus"][0]

        _mo_post = morahelper._mo_post
        _mo_post.return_value = attrdict(
            {
                "status_code": 201,
            }
        )
        self.assertFalse(_mo_post.called)

        sd_updater._create_engagement_type = MagicMock()
        sd_updater._create_engagement_type.return_value = "new_engagement_type_uuid"

        sd_updater._create_professions = MagicMock()
        sd_updater._create_professions.return_value = "new_profession_uuid"

        sd_updater.create_new_engagement(engagement, status, cpr, "user_uuid")
        _mo_post.assert_called_with(
            "details/create",
            {
                "type": "engagement",
                "org_unit": {"uuid": "department_uuid"},
                "person": {"uuid": "user_uuid"},
                "job_function": {"uuid": "new_profession_uuid"},
                "primary": {"uuid": "non_primary_uuid"},
                "engagement_type": {"uuid": engagement_type},
                "user_key": employment_id,
                "fraction": 0,
                "validity": {"from": "2020-11-10", "to": "2021-02-09"},
            },
        )
        sd_updater._create_engagement_type.assert_not_called()
        sd_updater._create_professions.assert_called_once()

    def test_terminate_engagement(self):

        employment_id = "01337"

        sd_updater = setup_sd_changed_at()
        morahelper = sd_updater.morahelper_mock

        sd_updater.mo_engagements_cache["user_uuid"] = [
            {
                "user_key": employment_id,
                "uuid": "mo_engagement_uuid",
            }
        ]

        _mo_post = morahelper._mo_post
        _mo_post.return_value = attrdict(
            {"status_code": 200, "text": lambda: "mo_engagement_uuid"}
        )
        self.assertFalse(_mo_post.called)
        sd_updater._terminate_engagement(
            user_key=employment_id, person_uuid="user_uuid", from_date="2020-11-01"
        )
        _mo_post.assert_called_once_with(
            "details/terminate",
            {
                "type": "engagement",
                "uuid": "mo_engagement_uuid",
                "validity": {"from": "2020-11-01", "to": None},
            },
        )

    def test_terminate_engagement_returns_false_when_no_mo_engagement(self):
        sd_updater = setup_sd_changed_at()

        morahelper = sd_updater.morahelper_mock
        mock_read_user_engagement = morahelper.read_user_engagement
        mock_read_user_engagement.return_value = []

        self.assertFalse(
            sd_updater._terminate_engagement(
                user_key="12345", person_uuid=str(uuid.uuid4()), from_date="2021-10-05"
            )
        )

    def test_terminate_engagement_when_to_date_set(self):
        sd_updater = setup_sd_changed_at()
        sd_updater.mo_engagements_cache["person_uuid"] = [
            {
                "user_key": "00000",
                "uuid": "mo_engagement_uuid",
            }
        ]

        morahelper = sd_updater.morahelper_mock
        mock_post_mo = morahelper._mo_post
        mock_post_mo.return_value = attrdict(
            {"status_code": 200, "text": lambda: "mo_engagement_uuid"}
        )

        sd_updater._terminate_engagement(
            user_key="00000",
            person_uuid="person_uuid",
            from_date="2021-10-15",
            to_date="2021-10-20",
        )
        mock_post_mo.assert_called_once_with(
            "details/terminate",
            {
                "type": "engagement",
                "uuid": "mo_engagement_uuid",
                "validity": {"from": "2021-10-15", "to": "2021-10-20"},
            },
        )

    @parameterized.expand([("2021-10-15", "2021-10-14"), ("9999-12-31", None)])
    def test_handle_status_changes_terminates_let_go_employment_status(
        self, sd_deactivation_date, mo_termination_to_date
    ):
        cpr = "0101709999"
        employment_id = "01337"

        _, read_employment_result = read_employment_fixture(
            cpr=cpr,
            employment_id=employment_id,
            job_id="1234",
            job_title="EDB-Mand",
            status="1",
        )
        sd_employment = read_employment_result[0]["Employment"]
        sd_employment["EmploymentStatus"].pop(0)  # Remove the one with status 8
        sd_employment["EmploymentStatus"][0]["DeactivationDate"] = sd_deactivation_date

        sd_updater = setup_sd_changed_at()
        sd_updater.mo_engagements_cache["person_uuid"] = [
            {
                "user_key": employment_id,
                "uuid": "mo_engagement_uuid",
            }
        ]

        morahelper = sd_updater.morahelper_mock
        mock_mo_post = morahelper._mo_post
        mock_mo_post.return_value = attrdict(
            {"status_code": 200, "text": lambda: "mo_engagement_uuid"}
        )

        sd_updater._handle_employment_status_changes(
            cpr=cpr, sd_employment=sd_employment, person_uuid="person_uuid"
        )

        mock_mo_post.assert_called_once_with(
            "details/terminate",
            {
                "type": "engagement",
                "uuid": "mo_engagement_uuid",
                "validity": {"from": "2021-02-10", "to": mo_termination_to_date},
            },
        )

    def test_handle_status_changes_terminates_slettet_employment_status(self):
        cpr = "0101709999"
        employment_id = "01337"

        _, read_employment_result = read_employment_fixture(
            cpr=cpr,
            employment_id=employment_id,
            job_id="1234",
            job_title="EDB-Mand",
            status="S",
        )
        sd_employment = read_employment_result[0]["Employment"]

        sd_updater = setup_sd_changed_at()
        sd_updater.mo_engagements_cache["person_uuid"] = [
            {
                "user_key": employment_id,
                "uuid": "mo_engagement_uuid",
            }
        ]

        morahelper = sd_updater.morahelper_mock
        mock_mo_post = morahelper._mo_post
        mock_mo_post.return_value = attrdict(
            {"status_code": 200, "text": lambda: "mo_engagement_uuid"}
        )

        sd_updater._handle_employment_status_changes(
            cpr=cpr, sd_employment=sd_employment, person_uuid="person_uuid"
        )

        mock_mo_post.assert_called_once_with(
            "details/terminate",
            {
                "type": "engagement",
                "uuid": "mo_engagement_uuid",
                "validity": {"from": "2020-11-01", "to": None},
            },
        )

    @parameterized.expand(
        [
            ["07777", "monthly pay"],
            ["90000", "hourly pay"],
            ["C3-P0", "employment pay"],
        ]
    )
    def test_update_all_employments_editing(self, employment_id, engagement_type):

        cpr = "0101709999"
        job_id = "1234"

        _, read_employment_result = read_employment_fixture(
            cpr=cpr,
            employment_id=employment_id,
            job_id=job_id,
            job_title="EDB-Mand",
        )

        sd_updater = setup_sd_changed_at()
        sd_updater.engagement_types = {
            "månedsløn": "monthly pay",
            "timeløn": "hourly pay",
            "engagement_type" + job_id: "employment pay",
        }

        sd_updater.read_employment_changed = lambda: read_employment_result
        # Load noop NY logic
        sd_updater.apply_NY_logic = (
            lambda org_unit, user_key, validity, person_uuid: org_unit
        )

        morahelper = sd_updater.morahelper_mock
        morahelper.read_user.return_value.__getitem__.return_value = "user_uuid"
        morahelper.read_user_engagement.return_value = [
            {
                "user_key": employment_id,
                "uuid": "mo_engagement_uuid",
                "validity": {"to": "9999-12-31"},
            }
        ]

        # Set primary types
        sd_updater.primary_types = {
            "primary": "primary_uuid",
            "non_primary": "non_primary_uuid",
            "no_salary": "no_salary_uuid",
            "fixed_primary": "fixed_primary_uuid",
        }

        _mo_post = morahelper._mo_post
        _mo_post.return_value = attrdict({"status_code": 201, "text": lambda: "OK"})
        self.assertFalse(_mo_post.called)

        sd_updater._create_engagement_type = MagicMock()
        sd_updater._create_engagement_type.return_value = "new_engagement_type_uuid"

        sd_updater._create_professions = MagicMock()
        sd_updater._create_professions.return_value = "new_profession_uuid"

        sd_updater.update_all_employments()
        # We expect the exact following 4 calls to have been made
        self.assertEqual(len(_mo_post.mock_calls), 5)
        _mo_post.assert_has_calls(
            [
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "mo_engagement_uuid",
                        "data": {
                            "validity": {"from": "2020-11-10", "to": "2021-02-09"},
                            "primary": {"uuid": "non_primary_uuid"},
                        },
                    },
                ),
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "mo_engagement_uuid",
                        "data": {
                            "org_unit": {"uuid": "department_uuid"},
                            "validity": {"from": "2020-11-10", "to": None},
                        },
                    },
                ),
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "mo_engagement_uuid",
                        "data": {
                            "job_function": {"uuid": "new_profession_uuid"},
                            "validity": {"from": "2020-11-10", "to": None},
                        },
                    },
                ),
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "mo_engagement_uuid",
                        "data": {
                            "engagement_type": {"uuid": engagement_type},
                            "validity": {"from": "2020-11-10", "to": None},
                        },
                    },
                ),
                call(
                    "details/terminate",
                    {
                        "type": "engagement",
                        "uuid": "mo_engagement_uuid",
                        "validity": {"from": "2021-02-10", "to": None},
                    },
                ),
            ]
        )
        sd_updater._create_engagement_type.assert_not_called()
        sd_updater._create_professions.assert_called_once()

    @given(job_position=st.integers(), no_salary_minimum=st.integers())
    @patch("integrations.SD_Lon.sd_changed_at.sd_payloads", autospec=True)
    def test_construct_object(self, sd_payloads_mock, job_position, no_salary_minimum):
        expected = no_salary_minimum is not None
        expected = expected and job_position < no_salary_minimum
        expected = not expected

        sd_updater = setup_sd_changed_at(
            {
                "sd_no_salary_minimum_id": no_salary_minimum,
            }
        )
        sd_updater.apply_NY_logic = (
            lambda org_unit, user_key, validity, person_uuid: org_unit
        )

        morahelper = sd_updater.morahelper_mock
        morahelper.read_ou.return_value = {
            "org_unit_level": {
                "user_key": "IHaveNoIdea",
            },
            "uuid": "uuid-a",
        }
        _mo_post = morahelper._mo_post
        _mo_post.return_value = attrdict({"status_code": 201, "text": lambda: "OK"})

        engagement = {
            "EmploymentIdentifier": "BIGAL",
            "EmploymentDepartment": [{"DepartmentUUIDIdentifier": "uuid-c"}],
            "Profession": [{"JobPositionIdentifier": str(job_position)}],
        }
        status = {
            "ActivationDate": "",
            "DeactivationDate": "",
            "EmploymentStatusCode": "",
        }
        cpr = ""
        result = sd_updater.create_new_engagement(engagement, status, cpr, "uuid-b")
        self.assertEqual(result, expected)
        if expected:
            sd_payloads_mock.create_engagement.assert_called_once()
        else:
            sd_payloads_mock.create_engagement.assert_not_called()

    @given(job_id=st.integers(min_value=0), engagement_exists=st.booleans())
    def test_fetch_engagement_type(self, job_id, engagement_exists):
        """Test that fetch_engagement_type only calls create when class is missing.

        This is done by creating the class if engagement_exists is set.
        I assume this works the same for _fetch_professions as they are similar.
        """
        sd_updater = setup_sd_changed_at()
        sd_updater.engagement_types = {
            "månedsløn": "monthly pay",
            "timeløn": "hourly pay",
        }

        self.assertEqual(len(sd_updater.engagement_types), 2)
        if engagement_exists:
            sd_updater.engagement_types[
                "engagement_type" + str(job_id)
            ] = "old_engagement_type_uuid"
            self.assertEqual(len(sd_updater.engagement_types), 3)
        else:
            self.assertEqual(len(sd_updater.engagement_types), 2)

        engagement_type_uuid = sd_updater._fetch_engagement_type(str(job_id))

        if engagement_exists:
            self.assertEqual(len(sd_updater.engagement_types), 3)
            sd_updater._create_class.assert_not_called()
            self.assertEqual(engagement_type_uuid, "old_engagement_type_uuid")
        else:
            self.assertEqual(len(sd_updater.engagement_types), 3)
            sd_updater._create_class.assert_called_once()
            sd_updater.job_sync.sync_from_sd.assert_called_once()
            self.assertIn("engagement_type" + str(job_id), sd_updater.engagement_types)
            self.assertEqual(engagement_type_uuid, "new_class_uuid")

    def test_edit_engagement(self):
        engagement = OrderedDict(
            [
                ("EmploymentIdentifier", "DEERE"),
                (
                    "Profession",
                    OrderedDict(
                        [
                            ("@changedAtDate", "1970-01-01"),
                            ("ActivationDate", "1960-01-01"),
                            ("DeactivationDate", "9999-12-31"),
                            ("JobPositionIdentifier", "9002"),
                            ("EmploymentName", "dummy"),
                            ("AppointmentCode", "0"),
                        ]
                    ),
                ),
            ]
        )

        sd_updater = setup_sd_changed_at()
        sd_updater.engagement_types = {
            "månedsløn": "monthly pay",
            "timeløn": "hourly pay",
        }
        sd_updater.mo_engagements_cache["person_uuid"] = [
            {
                "user_key": engagement["EmploymentIdentifier"],
                "uuid": "mo_engagement_uuid",
                "validity": {"from": "1950-01-01", "to": None},
            }
        ]

        morahelper = sd_updater.morahelper_mock
        _mo_post = morahelper._mo_post
        _mo_post.return_value = attrdict(
            {
                "status_code": 201,
            }
        )
        sd_updater._create_engagement_type = MagicMock(
            wraps=sd_updater._create_engagement_type
        )
        sd_updater._create_professions = MagicMock(wraps=sd_updater._create_professions)
        # Return 1 on first call, 2 on second call
        sd_updater._create_class.side_effect = [
            "new_class_1_uuid",
            "new_class_2_uuid",
        ]

        sd_updater.edit_engagement(engagement, "person_uuid")

        # Check that the create functions are both called
        sd_updater._create_engagement_type.assert_called_with(
            "engagement_type9002", "9002"
        )
        sd_updater._create_professions.assert_called_with("9002", "9002")
        # And thus that job_sync is called once from each
        sd_updater.job_sync.sync_from_sd.assert_has_calls(
            [call("9002", refresh=True), call("9002", refresh=True)]
        )
        # And that the results are returned to MO
        _mo_post.assert_has_calls(
            [
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "mo_engagement_uuid",
                        "data": {
                            "job_function": {"uuid": "new_class_1_uuid"},
                            "validity": {"from": "1960-01-01", "to": None},
                        },
                    },
                ),
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "mo_engagement_uuid",
                        "data": {
                            "engagement_type": {"uuid": "new_class_2_uuid"},
                            "validity": {"from": "1960-01-01", "to": None},
                        },
                    },
                ),
            ]
        )

    @patch("integrations.SD_Lon.sd_common.sd_lookup_settings")
    @patch("integrations.SD_Lon.sd_common.sd_lookup")
    def test_edit_engagement_job_position_id_set_to_value_above_9000(
        self, mock_sd_lookup, mock_sd_lookup_settings
    ):
        """
        If an employment exists in MO but with no engagement (e.g. which happens
        when the MO employment was created with an SD payload having a
        JobPositionIdentifier < 9000) and we receive an SD change payload, where
        the JobPositionIdentifier is set to a value greater than 9000, then we
        must ensure that an engagement is create for the corresponding employee
        in MO.
        """

        # Arrange

        sd_updater = setup_sd_changed_at(
            {
                "sd_monthly_hourly_divide": 80000,
                "sd_no_salary_minimum_id": 9000,
                "sd_import_too_deep": [
                    "Afdelings-niveau",
                    "NY1-niveau",
                ],
            }
        )

        sd_updater.engagement_types = {
            "månedsløn": "monthly pay",
            "timeløn": "hourly pay",
        }

        engagement = OrderedDict(
            [
                ("EmploymentIdentifier", "DEERE"),
                (
                    "Profession",
                    OrderedDict(
                        [
                            ("@changedAtDate", "2021-12-20"),
                            ("ActivationDate", "2021-12-19"),
                            ("DeactivationDate", "9999-12-31"),
                            ("JobPositionIdentifier", "9002"),
                            ("EmploymentName", "dummy"),
                            ("AppointmentCode", "0"),
                        ]
                    ),
                ),
            ]
        )
        mock_sd_lookup_settings.return_value = ("", "", "")

        # Necessary for the _find_engagement call in edit_engagement
        # Mock the call to arrange that no engagements are found for the user
        mora_helper = sd_updater.morahelper_mock
        _mo_lookup = mora_helper._mo_lookup
        _mo_lookup.return_value = []

        # Mock the call in sd_updater.read_employment_at(...)
        mock_sd_lookup.return_value = get_employment_fixture(
            1234561234, "emp_id", "dep_id", "dep_uuid", "9002", "job_title"
        )

        mock_apply_NY_logic = MagicMock()
        sd_updater.apply_NY_logic = mock_apply_NY_logic
        mock_apply_NY_logic.return_value = "org_unit_uuid"

        primary_types = sd_updater.primary_types_mock
        __getitem__ = primary_types.__getitem__
        __getitem__.return_value = "primary_uuid"

        _mo_post = mora_helper._mo_post
        _mo_post.return_value = attrdict(
            {
                "status_code": 201,
            }
        )

        # Act

        sd_updater.edit_engagement(engagement, "person_uuid")

        # Assert

        _mo_post.assert_called_once_with(
            "details/create",
            {
                "engagement_type": {"uuid": "new_class_uuid"},
                "fraction": 0,
                "job_function": {"uuid": "new_class_uuid"},
                "org_unit": {"uuid": "org_unit_uuid"},
                "person": {"uuid": "person_uuid"},
                "primary": {"uuid": "primary_uuid"},
                "type": "engagement",
                "user_key": "emp_id",
                "validity": {"from": "2020-11-10", "to": "2021-02-09"},
            },
        )

    @patch("integrations.SD_Lon.sd_common.sd_lookup_settings")
    def test_edit_engagement_profession_job_position_id_set_to_value_below_9000(
        self, mock_sd_lookup_settings
    ):
        """
        If an employment exists in MO WITH an engagement and we receive an
        SD change payload, where the JobPositionIdentifier is set to a value
        less than 9000, then we must ensure that an engagement is terminated
        for the corresponding employee in MO.
        """

        # Arrange

        sd_updater = setup_sd_changed_at(
            {
                "sd_monthly_hourly_divide": 80000,
                "sd_no_salary_minimum_id": 9000,
                "sd_import_too_deep": [
                    "Afdelings-niveau",
                    "NY1-niveau",
                ],
            }
        )

        engagement = OrderedDict(
            [
                ("EmploymentIdentifier", "DEERE"),
                (
                    "Profession",
                    OrderedDict(
                        [
                            ("@changedAtDate", "2021-12-20"),
                            ("ActivationDate", "2021-12-19"),
                            ("DeactivationDate", "9999-12-31"),
                            ("JobPositionIdentifier", "8000"),
                            ("EmploymentName", "dummy"),
                            ("AppointmentCode", "0"),
                        ]
                    ),
                ),
            ]
        )

        mo_eng = {"user_key": "12345", "person": {"uuid": "person_uuid"}}

        mock_sd_lookup_settings.return_value = ("", "", "")

        # Mock the terminate engagement call
        mock_terminate_engagement = MagicMock()
        sd_updater._terminate_engagement = mock_terminate_engagement
        mock_terminate_engagement.return_value = True

        # Act

        sd_updater.edit_engagement_profession(engagement, mo_eng)

        # Assert

        mock_terminate_engagement.assert_called_once_with(
            "12345", "person_uuid", "2021-12-19", None
        )

    @patch("integrations.SD_Lon.sd_changed_at.update_existing_engagement")
    def test_edit_engagement_handles_empty_professions_list(self, mock_update):
        """Handle an empty `professions` list in the engagement returned by SD"""
        # This is a regression test for #47799

        # Arrange
        sd_updater = setup_sd_changed_at()
        sd_updater._find_engagement = lambda *args: ["mo-eng"]
        engagement = OrderedDict(
            [
                ("EmploymentIdentifier", "DEERE"),
                ("Profession", []),
            ]
        )

        # Act
        sd_updater.edit_engagement(engagement, "person_uuid")

        # Assert
        mock_update.assert_called_once()

    @given(
        status=st.sampled_from(["1", "S"]),
        from_date=st.datetimes(),
        to_date=st.datetimes() | st.none(),
    )
    @patch("integrations.SD_Lon.sd_common.sd_lookup_settings")
    @patch("integrations.SD_Lon.sd_changed_at.sd_lookup")
    def test_timestamps_read_employment_changed(
        self,
        mock_sd_lookup,
        sd_settings,
        status,
        from_date,
        to_date,
    ):
        """Test that calls contain correct ActivationDate and ActivationTime"""
        sd_settings.return_value = ("", "", "")

        sd_updater = setup_sd_changed_at()
        sd_updater.read_employment_changed(from_date=from_date, to_date=to_date)
        expected_url = "GetEmploymentChangedAtDate20111201"
        url = mock_sd_lookup.call_args.args[0]
        params = mock_sd_lookup.call_args.kwargs["params"]
        self.assertEqual(url, expected_url)
        self.assertEqual(params["ActivationDate"], from_date.strftime("%d.%m.%Y"))
        self.assertEqual(params["ActivationTime"], from_date.strftime("%H:%M"))
        if to_date:
            self.assertEqual(params["DeactivationDate"], to_date.strftime("%d.%m.%Y"))
            self.assertEqual(params["DeactivationTime"], to_date.strftime("%H:%M"))

    @given(
        status=st.sampled_from(["1", "S"]),
        from_date=st.datetimes(),
        to_date=st.datetimes() | st.none(),
    )
    @patch("integrations.SD_Lon.sd_common.sd_lookup_settings")
    @patch("integrations.SD_Lon.sd_changed_at.sd_lookup")
    def test_timestamps_get_sd_persons_changed(
        self,
        mock_sd_lookup,
        sd_settings,
        status,
        from_date,
        to_date,
    ):
        """Test that calls contain correct ActivationDate and ActivationTime"""
        sd_settings.return_value = ("", "", "")

        sd_updater = setup_sd_changed_at()
        sd_updater.get_sd_persons_changed(from_date=from_date, to_date=to_date)
        expected_url = "GetPersonChangedAtDate20111201"
        url = mock_sd_lookup.call_args.args[0]
        params = mock_sd_lookup.call_args.kwargs["params"]
        self.assertEqual(url, expected_url)
        self.assertEqual(params["ActivationDate"], from_date.strftime("%d.%m.%Y"))
        self.assertEqual(params["ActivationTime"], from_date.strftime("%H:%M"))
        if to_date:
            self.assertEqual(params["DeactivationDate"], to_date.strftime("%d.%m.%Y"))
            self.assertEqual(params["DeactivationTime"], to_date.strftime("%H:%M"))
